from flask_wtf import FlaskForm
from wtforms import StringField, BooleanField
from wtforms.validators import Optional, DataRequired

from opentakserver.functions import false_values


class DataPackageUpdateForm(FlaskForm):
    hash = StringField(validators=[DataRequired()])
    install_on_enrollment = BooleanField(validators=[Optional()], false_values=false_values)
    install_on_connection = BooleanField(validators=[Optional()], false_values=false_values)



import datetime
import enum
import re
from dataclasses import dataclass

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.models.GroupUser import GroupUser
from sqlalchemy import Integer, String, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship


class GroupTypeEnum(str, enum.Enum):
    SYSTEM = "SYSTEM"
    LDAP = "LDAP"


class GroupDirectionEnum(str, enum.Enum):
    IN = "IN"  # Write
    OUT = "OUT"  # Read


@dataclass
class Group(db.Model):
    __tablename__ = "groups"

    IN = "IN"
    OUT = "OUT"
    SYSTEM = "SYSTEM"
    LDAP = "LDAP"

    def __init__(self):
        super().__init__()
        self.bitpos = self.get_next_bitpos()

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(String(255), unique=True)
    distinguishedName: Mapped[str] = mapped_column(String(255), nullable=True)
    created: Mapped[datetime] = mapped_column(DateTime, default=datetime.datetime.now(datetime.timezone.utc))
    type: Mapped[str] = mapped_column(String(255))
    bitpos: Mapped[int] = mapped_column(Integer)
    description: Mapped[str] = mapped_column(String, nullable=True)
    users = relationship("User", secondary="groups_users", viewonly=True, back_populates="groups", cascade="all, delete")
    missions = relationship("Mission", secondary="groups_missions", viewonly=True, back_populates="groups", cascade="all, delete")

    def get_next_bitpos(self) -> int:
        # the __ANON__ group is always 2 so default to 3 here
        bitpos = 3
        max_bitpos = db.session.execute(db.session.query(Group).order_by(Group.bitpos.desc()).limit(1)).first()
        if max_bitpos:
            bitpos = max_bitpos[0].bitpos + 1

        return bitpos

    def set_bitpos(self, new_bitpos : int = None):
        if new_bitpos:
            self.bitpos = new_bitpos
        else:
            self.bitpos = self.get_next_bitpos()

    def serialize(self):
        return {
            'name': self.name,
            'distinguishedName': self.distinguishedName,
            'created': self.created,
            'type': self.type,
            'bitpos': self.bitpos,
            'description': self.description,
        }

    def to_json(self):
        return_value = self.serialize()
        return_value['id'] = self.id
        return_value['bitpos'] = "{0:b}".format(self.bitpos)
        return_value['users'] = [user.serialize() for user in self.users]
        return_value['missions'] = [mission.serialize() for mission in self.missions]
        return return_value

    def to_marti_json_in(self):
        # Remove the _READ and _WRITE suffixes when using LDAP for groups
        remove_read = re.compile(re.escape("_read"), re.IGNORECASE)
        remove_write = re.compile(re.escape("_write"), re.IGNORECASE)
        group_name = remove_read.sub("", self.name)
        group_name = remove_write.sub("", group_name)

        return {
            'name': group_name,
            'direction': Group.IN,
            'created': iso8601_string_from_datetime(self.created or datetime.datetime.now(tz=datetime.timezone.utc)).split("T")[0],
            'type': self.type,
            'bitpos': self.bitpos,
            'active': True,
            'description': self.description or "",
        }

    def to_marti_json_out(self):
        return_value = self.to_marti_json_in()
        return_value['direction'] = Group.OUT
        return return_value



import traceback

from flask import Blueprint, request, jsonify
from flask_babel import gettext
from flask_security import auth_required, roles_required
from sqlalchemy import update
from sqlalchemy.exc import IntegrityError

from opentakserver.extensions import db, logger
from opentakserver.forms.device_profile_form import DeviceProfileForm
from opentakserver.models.DeviceProfiles import DeviceProfiles
from opentakserver.blueprints.ots_api.api import search, paginate

device_profile_api_blueprint = Blueprint('device_profile_api_blueprint', __name__)


@device_profile_api_blueprint.route('/api/profiles')
@roles_required("administrator")
def get_device_profiles():
    query = db.session.query(DeviceProfiles)
    query = search(query, DeviceProfiles, 'preference_key')
    query = search(query, DeviceProfiles, 'preference_value')
    query = search(query, DeviceProfiles, 'tool')
    return paginate(query)


@device_profile_api_blueprint.route('/api/profiles', methods=['POST'])
@auth_required()
@roles_required("administrator")
def add_device_profile():
    form = DeviceProfileForm()
    if not form.validate():
        return jsonify({'success': False, 'errors': form.errors}), 400

    device_profile = DeviceProfiles()
    device_profile.from_wtf(form)

    try:
        db.session.add(device_profile)
        db.session.commit()
    except IntegrityError:
        db.session.rollback()
        db.session.execute(update(DeviceProfiles).where(DeviceProfiles.preference_key == device_profile.preference_key)
                           .values(**device_profile.serialize()))
        db.session.commit()

    return jsonify({'success': True})


@device_profile_api_blueprint.route('/api/profiles', methods=['DELETE'])
@roles_required("administrator")
def delete_device_profile():
    preference_key = request.args.get('preference_key')
    if not preference_key:
        return jsonify({'success': False, 'error': gettext(u'Please specify the preference_key')}), 400
    try:
        query = db.session.query(DeviceProfiles)
        query = search(query, DeviceProfiles, 'preference_key')
        preference = db.session.execute(query).first()
        if not preference:
            return jsonify({'success': False, 'error': gettext(u'Unknown preference_key: %(preference_key)s', preference_key=preference_key)}), 404

        db.session.delete(preference[0])
        db.session.commit()
        return jsonify({'success': True})
    except BaseException as e:
        logger.error(f"Failed to delete device profile: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 400



"""Fixes for the packages table, added group_eud table

Revision ID: f107b45529ba
Revises: 7e9f5d2c193d
Create Date: 2024-10-28 14:36:54.189219

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'f107b45529ba'
down_revision = '7e9f5d2c193d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('groups_euds',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('group_id', sa.Integer(), nullable=False),
    sa.Column('direction', sa.String(length=255), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=False),
    sa.Column('description', sa.String(length=255), nullable=True),
    sa.Column('eud_uid', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['eud_uid'], ['euds.uid'], name="group_eud"),
    sa.ForeignKeyConstraint(['group_id'], ['groups.id'], name="group"),
    sa.PrimaryKeyConstraint('id')
    )

    with op.batch_alter_table('groups', schema=None) as batch_op:
        batch_op.drop_constraint('eud', type_='foreignkey')
        batch_op.drop_column('description')
        batch_op.drop_column('active')
        batch_op.drop_column('direction')
        batch_op.drop_column('eud_uid')

    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.alter_column('description',
               existing_type=sa.Integer(),
               type_=sa.String(length=255),
               existing_nullable=True)
        batch_op.alter_column('apk_hash',
               existing_type=sa.Integer(),
               type_=sa.String(length=255),
               existing_nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.alter_column('apk_hash',
               existing_type=sa.String(length=255),
               type_=sa.Integer(),
               existing_nullable=True)
        batch_op.alter_column('description',
               existing_type=sa.String(length=255),
               type_=sa.Integer(),
               existing_nullable=True)

    with op.batch_alter_table('groups', schema=None) as batch_op:
        batch_op.add_column(sa.Column('eud_uid', sa.String(length=255), nullable=False))
        batch_op.add_column(sa.Column('direction', sa.String(length=255), nullable=False))
        batch_op.add_column(sa.Column('active', sa.Boolean(), autoincrement=False, nullable=False))
        batch_op.add_column(sa.Column('description', sa.String(length=255), nullable=True))
        batch_op.create_foreign_key('eud', 'euds', ['eud_uid'], ['uid'])

    op.drop_table('groups_euds')
    # ### end Alembic commands ###



from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, JSON, ForeignKey, DateTime, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


class CoT(db.Model):
    __tablename__ = "cot"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    how: Mapped[str] = mapped_column(String(255), nullable=True)
    type: Mapped[str] = mapped_column(String(255), nullable=True)
    uid: Mapped[str] = mapped_column(String(255), nullable=True)
    sender_callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    sender_device_name: Mapped[str] = mapped_column(String(255), nullable=True)
    sender_uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True)
    recipients: Mapped[JSON] = mapped_column(JSON, nullable=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime)
    start: Mapped[datetime] = mapped_column(DateTime)
    stale: Mapped[datetime] = mapped_column(DateTime)
    xml: Mapped[str] = mapped_column(TEXT)
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey("missions.name"), nullable=True)
    eud = relationship("EUD", back_populates="cots", uselist=False)
    alert = relationship("Alert", cascade="all, delete, delete-orphan", back_populates="cot", uselist=False)
    point = relationship("Point", cascade="all, delete, delete-orphan", back_populates="cot", uselist=False)
    casevac = relationship("CasEvac", cascade="all, delete, delete-orphan", back_populates="cot", uselist=False)
    video = relationship("VideoStream", back_populates="cot", cascade="all, delete, delete-orphan", uselist=False)
    geochat = relationship("GeoChat", back_populates="cot", cascade="all, delete, delete-orphan", uselist=False)
    marker = relationship("Marker", back_populates="cot", cascade="all, delete, delete-orphan", uselist=False)
    rb_line = relationship("RBLine", cascade="all, delete, delete-orphan", back_populates="cot")
    mission = relationship("Mission", back_populates="cots")

    def serialize(self):
        return {
            'how': self.how,
            'type': self.type,
            'uid': self.uid,
            'sender_callsign': self.sender_callsign,
            'sender_uid': self.sender_uid,
            'recipients': self.recipients,
            'timestamp': self.timestamp,
            'start': self.start,
            'stale': self.stale,
            'xml': self.xml,
        }

    def to_json(self):
        return {
            'how': self.how,
            'type': self.type,
            'uid': self.uid,
            'sender_callsign': self.sender_callsign,
            'sender_uid': self.sender_uid,
            'recipients': self.recipients,
            'timestamp': self.timestamp,
            'start': iso8601_string_from_datetime(self.start),
            'stale': iso8601_string_from_datetime(self.stale),
            'xml': self.xml,
            'eud': self.eud.to_json() if self.eud else None,
            'alert': self.alert.to_json() if self.alert else None,
            'point': self.point.to_json() if self.point else None,
            'casevac': self.casevac.to_json() if self.casevac else None,
            'video': self.video.to_json() if self.video else None,
            'geochat': self.geochat.to_json() if self.geochat else None,
        }



import os
import platform
import sys

import click
import yaml
from flask import g, current_app as app
from flask.cli import with_appcontext

from opentakserver.certificate_authority import CertificateAuthority
from opentakserver.defaultconfig import DefaultConfig
from opentakserver.extensions import logger


@click.group()
@click.option('-x', '--x-arg', multiple=True,
              help='Additional arguments consumed by custom env.py scripts')
@with_appcontext
def ots(x_arg):
    """Create an OpenTAKServer CA and config file"""
    g.x_arg = x_arg


@ots.command()
@with_appcontext
def create_ca():
    ca = CertificateAuthority(logger, app)
    if not ca.check_if_ca_exists():
        logger.info("Creating certificate authority...")
        ca.create_ca()
    else:
        logger.warning("Certificate authority already exists")
    sys.exit()


@ots.command()
@click.option("--overwrite", is_flag=True)
@with_appcontext
def generate_config(overwrite):
    app.config.from_object(DefaultConfig)

    if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")) and not overwrite:
        logger.warning("config.yml already exists")
        return

    logger.info("Creating config.yml")
    with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "w") as config:
        conf = {}
        for option in DefaultConfig.__dict__:
            # Fix the sqlite DB path on Windows
            if option == "SQLALCHEMY_DATABASE_URI" and platform.system() == "Windows" and DefaultConfig.__dict__[option].startswith("sqlite"):
                conf[option] = DefaultConfig.__dict__[option].replace("////", "///").replace("\\", "/")
            elif option.isupper():
                conf[option] = DefaultConfig.__dict__[option]
        config.write(yaml.safe_dump(conf))


@click.group()
@click.option('-x', '--x-arg', multiple=True,
              help='Additional arguments consumed by custom env.py scripts')
@with_appcontext
def translate(x_arg):
    """Create new translations"""
    g.x_arg = x_arg


@translate.command()
@with_appcontext
def update():
    if os.system('pybabel extract -F babel.cfg -k _l -o messages.pot .'):
        raise RuntimeError('extract command failed')
    if os.system('pybabel update -i messages.pot -d translations'):
        raise RuntimeError('update command failed')
    os.remove('messages.pot')


@translate.command()
@with_appcontext
def compile():
    if os.system('pybabel compile -d translations'):
        raise RuntimeError('compile command failed')


@translate.command()
@click.argument('lang')
def init(lang):
    """Initialize a new language."""
    if os.system('pybabel extract -F babel.cfg -k _l -o messages.pot .'):
        raise RuntimeError('extract command failed')
    if os.system(
            'pybabel init -i messages.pot -d translations -l ' + lang):
        raise RuntimeError('init command failed')
    os.remove('messages.pot')



"""Added uid column to cot table

Revision ID: 00546817d518
Revises: 807c3ca8e7d0
Create Date: 2024-10-11 20:12:20.866452

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '00546817d518'
down_revision = '807c3ca8e7d0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.add_column(sa.Column('uid', sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.drop_column('uid')

    # ### end Alembic commands ###



import os
import re
import traceback
from datetime import datetime, timezone

from flask import Blueprint, request, jsonify, current_app as app
from flask_security import roles_required

import httpx

from opentakserver import __version__ as version
from opentakserver.blueprints.ots_api.api import change_config_setting
from opentakserver.blueprints.ots_api.package_api import create_product_infz
from opentakserver.extensions import logger, db
from opentakserver.models.Packages import Packages

tak_gov_link_blueprint = Blueprint('tak_gov_link_blueprint', __name__)

HEADERS = {"User-Agent": f"OpenTAKServer {version}"}


def get_new_access_token():
    try:
        client = httpx.Client(http2=True)
        refresh_payload = {"client_id": "tak-gov-eud", "grant_type": "refresh_token", "refresh_token": app.config.get("OTS_TAK_GOV_REFRESH_TOKEN")}
        response = client.post("https://auth.tak.gov/auth/realms/TPC/protocol/openid-connect/token", data=refresh_payload, headers=HEADERS)

        if response.status_code != 200:
            logger.error({"success": False, "error": f"Failed to get new access token: {response.text}"})
            return {"success": False, "error": f"Failed to get new access token: {response.text}"}

        response_data = response.json()
        access_token = response_data["access_token"]
        refresh_token = response_data["refresh_token"]
        expires_in = response_data["expires_in"]
        change_config_setting("OTS_TAK_GOV_REFRESH_TOKEN", refresh_token)
        app.config["OTS_TAK_GOV_REFRESH_TOKEN"] = refresh_token

        return {"success": True, "access_token": access_token, "expires_in": expires_in}
    except BaseException as e:
        logger.error(f"Failed to get new access token: {e}")
        logger.debug(traceback.format_exc())
        return {"success": False, "error": f"Failed to get new access token: {e}"}


@tak_gov_link_blueprint.route('/api/takgov/link')
@roles_required('administrator')
def link_tak_gov_account():
    """Interacts with tak.gov's eud_api to get the user_code and device_code required to link OpenTAKServer to a tak.gov account

    :return: The device_code and user_code on success, error message otherwise
    """

    try:
        # Use httpx since it supports http/2.0
        client = httpx.Client(http2=True)

        # Step 1: Get the device_code and user_code
        payload = {"client_id": "tak-gov-eud", "scope": "openid offline_access email profile", "grant_type": "urn:ietf:params:oauth:grant-type:device_code"}
        response = client.post("https://auth.tak.gov/auth/realms/TPC/protocol/openid-connect/auth/device", headers=HEADERS, data=payload)
        if response.status_code != 200:
            return jsonify({"success": False, "error": f"Failed to link tak.gov account: {response.text}"})

        response_data = response.json()
        device_code = response_data["device_code"]
        verification_uri = response_data["verification_uri"]
        verification_uri_complete = response_data["verification_uri_complete"]
        user_code = response_data["user_code"]
        expires_in = response_data["expires_in"]
        interval = response_data["interval"]

        return jsonify({"success": True, "device_code": device_code, "user_code": user_code})

    except BaseException as e:
        logger.error(f"Failed to link tak.gov account: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": f"Failed to link tak.gov account: {e}"}), 500


@tak_gov_link_blueprint.route('/api/takgov/token')
@roles_required('administrator')
def get_initial_tak_gov_token():
    """Queries the tak.gov eud_api to get the initial refresh_token and access_token once the user has entered their user_code into tak.gov.

    :return: JSON object with access_token and expires_in, error message otherwise
    """

    device_code = request.args.get('device_code')
    if not device_code:
        return jsonify({"success": False, "error": "device_code is required"}), 400

    try:
        client = httpx.Client(http2=True)

        payload = {"grant_type": "urn:ietf:params:oauth:grant-type:device_code", "device_code": device_code, "client_id": "tak-gov-eud"}
        r = client.post("https://auth.tak.gov/auth/realms/TPC/protocol/openid-connect/token", data=payload, headers=HEADERS)

        logger.info(r.text)
        response_data = r.json()

        access_token = response_data["access_token"]
        refresh_token = response_data["refresh_token"]
        expires_in = response_data["expires_in"]

        change_config_setting("OTS_TAK_GOV_LINKED", True)
        change_config_setting("OTS_TAK_GOV_REFRESH_TOKEN", refresh_token)

        app.config.update({"OTS_TAK_GOV_LINKED": True})
        app.config.update({"OTS_TAK_GOV_REFRESH_TOKEN": refresh_token})

        return jsonify({"success": True, "access_token": access_token, "expires_in": expires_in})
    except BaseException as e:
        logger.error(f"Failed to get initial tak gov token: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": f"Failed to get initial tak gov token: {e}"})


@tak_gov_link_blueprint.route('/api/takgov/token', methods=["PATCH"])
@roles_required('administrator')
def get_new_tak_gov_token():
    """Uses the refresh_token to get a new access_token. This is required since the access_token expires three minutes after
    it is generated.

    :return: JSON object with the new access_token, error message otherwise.
    """

    if not app.config.get("OTS_TAK_GOV_LINKED") or not app.config.get("OTS_TAK_GOV_REFRESH_TOKEN"):
        return jsonify({"success": False, "error": "Not linked to a tak.gov account"}), 400

    try:
        token = get_new_access_token()
        if token["success"]:
            return jsonify(token)
        else:
            return jsonify(token), 500

    except BaseException as e:
        logger.error(f"Failed to get new tak.gov token: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": f"Failed to get new tak.gov token {e}"}), 500


@tak_gov_link_blueprint.route('/api/takgov/plugins')
@roles_required('administrator')
def get_plugins_list():
    """Query tak.gov's eud_api for a list of plugins for the specified product and product version (i.e. ATAK-CIV 5.5.0)

    :return: JSON array of available plugins.
    """
    product = request.args.get('product')
    product_version = request.args.get('product_version')

    if not product or not product_version:
        return jsonify({"success": False, "error": "product and product_version are required"}), 400

    token = get_new_access_token()
    if not token["success"]:
        return jsonify(token), 500

    client = httpx.Client(http2=True)
    HEADERS["Authorization"] = f"Bearer {token['access_token']}"
    params = {"product": product, "product_version": product_version}

    try:
        response = client.get("https://tak.gov/eud_api/software/v1/plugins", params=params, headers=HEADERS)
        return jsonify(response.json())
    except BaseException as e:
        logger.error(f"Failed to get plugin list: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": f"Failed to get plugin list: {e}"}), 500


@tak_gov_link_blueprint.route('/api/takgov', methods=['DELETE'])
@roles_required("administrator")
def unlink_tak_gov_account():
    if not app.config.get('OTS_TAK_GOV_LINKED'):
        return jsonify({"success": False, "error": "Not linked to a TAK.gov account"}), 400

    change_config_setting("OTS_TAK_GOV_LINKED", False)
    change_config_setting("OTS_TAK_GOV_REFRESH_TOKEN", None)

    app.config.update({"OTS_TAK_GOV_LINKED": False})
    app.config.update({"OTS_TAK_GOV_REFRESH_TOKEN": None})

    return jsonify({"success": True})


@tak_gov_link_blueprint.route('/api/takgov')
@roles_required("administrator")
def check_if_linked():
    return jsonify({"success": True, "tak_gov_account_linked": app.config.get('OTS_TAK_GOV_LINKED')})


@tak_gov_link_blueprint.route('/api/takgov/icon')
@roles_required("administrator")
def get_plugin_icon():
    """Downloads the plugin's icon from tak.gov's eud_api using the access_token

    :return: The icon or an error message.
    """
    icon_url = request.args.get('icon_url')
    if not icon_url:
        return jsonify({"success": False, "error": "icon_link is required"}), 400
    elif not icon_url.startswith("https://tak.gov/eud_api"):
        return jsonify({"success": False, "error": "icon_link is invalid"}), 400

    token = get_new_access_token()
    if not token["success"]:
        return jsonify(token), 500

    client = httpx.Client(http2=True)
    HEADERS["Authorization"] = f"Bearer {token['access_token']}"
    response = client.get(icon_url, headers=HEADERS)
    logger.warning(response.content)
    logger.warning(response.text)
    logger.warning(response.headers)

    return response.content


@tak_gov_link_blueprint.route('/api/takgov/plugin', methods=["POST"])
@roles_required("administrator")
def download_plugin():
    apk_hash = request.json.get('apk_hash')
    apk_size = request.json.get('apk_size_bytes')
    apk_url = request.json.get('apk_url')
    apk_type = request.json.get('apk_type')
    description = request.json.get('description')
    name = request.json.get('display_name')
    platform = request.json.get('platform')
    package_name = request.json.get('package_name')
    plugin_version = request.json.get('version')
    revision_code = request.json.get('revision_code')
    os_requirement = request.json.get('os_requirement')
    tak_prerequisite = request.json.get('tak_prerequisite')
    atak_version = request.json.get('atak_version')

    # Don't set an ATAK version for the plugin for ATAK < 5.5.0.
    # ATAK 5.4.0 and below don't check for plugins for their specific version of ATAK
    major, minor, patch = atak_version.split(".")
    if int(major) < 5 or (int(major) == 5 and int(minor) < 5):
        atak_version = None

    if not apk_url:
        return jsonify({"success": False, "error": "plugin_url is required"}), 400

    exising_plugin = db.session.execute(db.session.query(Packages).filter_by(version=plugin_version, package_name=package_name)).first()
    if exising_plugin:
        return jsonify({"success": False, "error": f"Plugin {package_name} already exists"}), 400

    client = httpx.Client(http2=True)
    token = get_new_access_token()
    if not token["success"]:
        return jsonify(token), 500

    HEADERS["Authorization"] = f"Bearer {token['access_token']}"
    response = client.get(apk_url, headers=HEADERS, follow_redirects=True)
    if response.status_code != 200:
        return jsonify(response.content), 500

    content_disposition = response.headers.get("content-disposition")
    filename = re.search(r'filename=\"(.+)\"', content_disposition).group(1)
    with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", filename), "wb") as f:
        f.write(response.content)

    icon = None
    icon_filename = None
    icon_response = client.get(request.json.get('icon_url'), headers=HEADERS, follow_redirects=True)
    # Some plugins don't have icons and will return a 404
    if icon_response.status_code == 200:
        icon = icon_response.content
        content_disposition = icon_response.headers.get("content-disposition")
        icon_filename = re.search(r'filename=\"(.+)\"', content_disposition).group(1)

    plugin = Packages()
    plugin.platform = platform
    plugin.apk_hash = apk_hash
    plugin.file_size = apk_size
    plugin.package_name = package_name
    plugin.version = plugin_version
    plugin.description = description
    plugin.name = name
    plugin.revision_code = revision_code
    plugin.os_requirement = os_requirement
    plugin.plugin_type = apk_type
    plugin.atak_version = atak_version
    plugin.icon = icon
    plugin.icon_filename = icon_filename
    plugin.file_name = filename
    plugin.publish_time = datetime.now(tz=timezone.utc)

    try:
        db.session.add(plugin)
        db.session.commit()
    except BaseException as e:
        logger.error(f"Failed to add plugin to database: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": f"Failed to add plugin to database: {e}"}), 500

    create_product_infz(atak_version)

    return jsonify({"success": True})



"""Added eud_stats table

Revision ID: 1041fae84708
Revises: 21fb5a21f356
Create Date: 2024-12-17 04:37:54.374451

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1041fae84708'
down_revision = '21fb5a21f356'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('eud_stats',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('timestamp', sa.DateTime(), nullable=True),
    sa.Column('eud_uid', sa.String(length=255), nullable=False),
    sa.Column('heap_free_size', sa.Integer(), nullable=True),
    sa.Column('app_framerate', sa.Integer(), nullable=True),
    sa.Column('storage_total', sa.Integer(), nullable=True),
    sa.Column('heap_current_size', sa.Integer(), nullable=True),
    sa.Column('battery', sa.Integer(), nullable=True),
    sa.Column('battery_temp', sa.Integer(), nullable=True),
    sa.Column('deviceDataRx', sa.Integer(), nullable=True),
    sa.Column('heap_max_size', sa.Integer(), nullable=True),
    sa.Column('storage_available', sa.Integer(), nullable=True),
    sa.Column('deviceDataTx', sa.Integer(), nullable=True),
    sa.Column('ip_address', sa.String(length=255), nullable=True),
    sa.Column('battery_status', sa.String(length=255), nullable=True),
    sa.ForeignKeyConstraint(['eud_uid'], ['euds.uid'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )

def downgrade():
    op.drop_table('eud_stats')
    # ### end Alembic commands ###



"""Fixed video_streams table

Revision ID: 298430e7849d
Revises: f107b45529ba
Create Date: 2024-10-28 19:14:27.017319

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '298430e7849d'
down_revision = 'f107b45529ba'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('video_streams', schema=None) as batch_op:
        batch_op.alter_column('xml',
               existing_type=sa.VARCHAR(length=255),
               type_=sa.TEXT(),
               existing_nullable=True)
        batch_op.alter_column('mediamtx_settings',
               existing_type=sa.VARCHAR(length=255),
               type_=sa.TEXT(),
               existing_nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('video_streams', schema=None) as batch_op:
        batch_op.alter_column('mediamtx_settings',
               existing_type=sa.TEXT(),
               type_=sa.VARCHAR(length=255),
               existing_nullable=False)
        batch_op.alter_column('xml',
               existing_type=sa.TEXT(),
               type_=sa.VARCHAR(length=255),
               existing_nullable=True)

    # ### end Alembic commands ###



from bs4 import BeautifulSoup
from androguard.core.apk import APK

from sqlalchemy import Integer, String, Boolean
from sqlalchemy.orm import Mapped, mapped_column

from opentakserver.extensions import db


class Plugins(db.Model):
    __tablename__ = "plugins"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(String(255), unique=True)
    distro: Mapped[str] = mapped_column(String(255), unique=True)
    enabled: Mapped[bool] = mapped_column(Boolean, default=True)
    author: Mapped[str] = mapped_column(String(255), nullable=True, default=None)
    version: Mapped[str] = mapped_column(String(255), nullable=True, default=None)



import datetime
from dataclasses import dataclass

from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.extensions import db
from sqlalchemy import String, DateTime, Float, ForeignKey, Integer
from sqlalchemy.orm import Mapped, mapped_column, relationship


@dataclass
class MissionUID(db.Model):
    __tablename__ = "mission_uids"

    uid: Mapped[str] = mapped_column(String(255), primary_key=True)  # Equals the original CoT's UID
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey("missions.name"), nullable=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    creator_uid: Mapped[str] = mapped_column(String(255), nullable=True)
    cot_type: Mapped[str] = mapped_column(String(255), nullable=True)
    callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    iconset_path: Mapped[str] = mapped_column(String(255), nullable=True)
    color: Mapped[str] = mapped_column(String(255), nullable=True)
    latitude: Mapped[float] = mapped_column(Float, nullable=True)
    longitude: Mapped[float] = mapped_column(Float, nullable=True)
    mission = relationship("Mission", back_populates="uids")
    mission_change = relationship("MissionChange", back_populates="uid", uselist=False)

    def serialize(self):
        return {
            'timestamp': self.timestamp,
            'creator_uid': self.creator_uid,
            'cot_type': self.cot_type,
            'callsign': self.callsign,
            'iconset_path': self.iconset_path,
            'color': self.color,
            'latitude': self.latitude,
            'longitude': self.longitude
        }

    def to_json(self):
        return {
            'data': self.uid,
            'timestamp': iso8601_string_from_datetime(self.timestamp),
            'creatorUid': self.creator_uid if self.creator_uid else "",
            'details': {
                'type': self.cot_type,
                'callsign': self.callsign,
                'iconsetPath': self.iconset_path,
                'color': self.color,
                'location': {
                    'lat': self.latitude,
                    'lon': self.longitude
                }
            }
        }

    def to_details_json(self):
        return {
            'type': self.cot_type,
            'callsign': self.callsign,
            'iconsetPath': self.iconset_path,
            'color': self.color,
            'location': {
                'lat': self.latitude,
                'lon': self.longitude
            }
        }



import json
import traceback
from urllib.parse import urlparse
from xml.etree.ElementTree import tostring, fromstring, Element

import sqlalchemy
from OpenSSL import crypto
from bs4 import BeautifulSoup
from flask import request, Blueprint, jsonify, current_app as app
from flask_babel import gettext
from werkzeug.datastructures import ImmutableMultiDict

from opentakserver.blueprints.marti_api.marti_api import verify_client_cert
from opentakserver.extensions import logger, db
from opentakserver.forms.MediaMTXPathConfig import MediaMTXPathConfig
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver import __version__ as version
from opentakserver.models.EUD import EUD
from opentakserver.models.VideoStream import VideoStream
from opentakserver.models.user import User

video_marti_api = Blueprint('video_marti_api', __name__)

@video_marti_api.route('/Marti/vcm', methods=['GET', 'POST'])
def video():
    if request.method == 'POST':
        soup = BeautifulSoup(request.data, 'xml')
        video_connections = soup.find('videoConnections')

        path = video_connections.find('path').text
        if path.startswith("/"):
            path = path[1:]

        if video_connections:
            v = VideoStream()
            v.protocol = video_connections.find('protocol').text
            v.alias = video_connections.find('alias').text
            v.uid = video_connections.find('uid').text
            v.port = video_connections.find('port').text
            v.rover_port = video_connections.find('roverPort').text
            v.ignore_embedded_klv = (video_connections.find('ignoreEmbeddedKLV').text.lower() == 'true')
            v.preferred_mac_address = video_connections.find('preferredMacAddress').text
            v.preferred_interface_address = video_connections.find('preferredInterfaceAddress').text
            v.path = path
            v.buffer_time = video_connections.find('buffer').text
            v.network_timeout = video_connections.find('timeout').text
            v.rtsp_reliable = video_connections.find('rtspReliable').text
            path_config = MediaMTXPathConfig(None).serialize()
            path_config['sourceOnDemand'] = False
            v.mediamtx_settings = json.dumps(path_config)

            # Discard username and password for security
            feed = soup.find('feed')
            address = feed.find('address').text
            feed.find('address').string.replace_with(address.split("@")[-1])

            v.xml = str(feed)

            with app.app_context():
                try:
                    db.session.add(v)
                    db.session.commit()
                    logger.debug("Inserted Video")
                except sqlalchemy.exc.IntegrityError as e:
                    db.session.rollback()
                    v = db.session.execute(db.select(VideoStream).filter_by(path=v.path)).scalar_one()
                    v.protocol = video_connections.find('protocol').text
                    v.alias = video_connections.find('alias').text
                    v.uid = video_connections.find('uid').text
                    v.port = video_connections.find('port').text
                    v.rover_port = video_connections.find('roverPort').text
                    v.ignore_embedded_klv = (video_connections.find('ignoreEmbeddedKLV').text.lower() == 'true')
                    v.preferred_mac_address = video_connections.find('preferredMacAddress').text
                    v.preferred_interface_address = video_connections.find('preferredInterfaceAddress').text
                    v.path = video_connections.find('path').text
                    v.buffer_time = video_connections.find('buffer').text
                    v.network_timeout = video_connections.find('timeout').text
                    v.rtsp_reliable = video_connections.find('rtspReliable').text
                    feed = soup.find('feed')
                    address = feed.find('address').text
                    feed.find('address').replace_with(address.split("@")[-1])

                    v.xml = str(feed)

                    db.session.commit()
                    logger.debug("Updated video")

        return '', 200

    elif request.method == 'GET':
        try:
            with app.app_context():
                videos = db.session.execute(db.select(VideoStream)).scalars()
                videoconnections = Element('videoConnections')

                for video in videos:
                    # Make sure videos have the correct address based off of the Flask request and not 127.0.0.1
                    # This also forces all streams to bounce through MediaMTX
                    feed = BeautifulSoup(video.xml, 'xml')

                    url = urlparse(request.url_root).hostname
                    path = feed.find('path').text
                    if not path.startswith("/"):
                        path = "/" + path

                    if 'iTAK' in request.user_agent.string:
                        url = feed.find('protocol').text + "://" + url + ":" + feed.find("port").text + path

                    if feed.find('address'):
                        feed.find('address').string.replace_with(url)
                    else:
                        address = feed.new_tag('address')
                        address.string = url
                        feed.feed.append(address)
                    videoconnections.append(fromstring(str(feed)))

            return tostring(videoconnections), 200
        except BaseException as e:
            logger.error(traceback.format_exc())
            return '', 500


@video_marti_api.route('/Marti/api/video')
def get_videos():
    cert = verify_client_cert()
    if not cert:
        # Shouldn't ever get here since nginx already verifies the cert
        return jsonify({'success': False, 'error': gettext(u'Invalid Certificate')}), 400
    username = None
    for a in cert.get_subject().get_components():
        if a[0].decode('UTF-8') == 'CN':
            username = a[1].decode('UTF-8')
            break

    user = app.security.datastore.find_user(username=username)
    videos = db.session.execute(db.select(VideoStream)).scalars()

    video_connections = {'videoConnections': []}
    for video in videos:
        video_connections['videoConnections'].append(video.to_marti_json(user))

    return jsonify(video_connections)


@video_marti_api.route('/Marti/api/video', methods=['POST'])
def add_video():
    cert = verify_client_cert()
    username = None
    for a in cert.get_subject().get_components():
        if a[0].decode('UTF-8') == 'CN':
            username = a[1].decode('UTF-8')
            break

    videos = request.json

    for video in videos['videoConnections']:
        feeds = video.get('feeds')

        for feed in feeds:
            data = ImmutableMultiDict(
                {
                    'path': feed['alias'],
                    'source': feed['url']
                }
            )
            mediamtx_config = MediaMTXPathConfig(formdata=data, csrf_enabled=False)

            video_stream = VideoStream()
            video_stream.path = feed['alias']
            video_stream.uid = video['uuid']
            video_stream.mediamtx_settings = json.dumps(mediamtx_config.serialize())
            video_stream.username = username
            video_stream.rover_port = -1
            video_stream.ignore_embedded_klv = False
            video_stream.buffer_time = None
            video_stream.rtsp_reliable = 1
            video_stream.network_timeout = 10000
            video_stream.generate_xml(urlparse(request.url_root).hostname)
            db.session.add(video_stream)
            db.session.commit()

    return '', 200


@video_marti_api.route('/Marti/api/video/<uid>')
def get_video(uid):
    cert = verify_client_cert()
    username = None
    for a in cert.get_subject().get_components():
        if a[0].decode('UTF-8') == 'CN':
            username = a[1].decode('UTF-8')
            break

    video = db.session.query(VideoStream).filter_by(uid=uid).scalar()
    if not video:
        return jsonify({'success': False, 'error': gettext(u'Video not found')}), 404
    user = db.session.execute(db.session.query(User).filter_by(username=username)).first()
    if not user:
        return jsonify({'success': False, 'error': gettext(u'User %(username)s not found', username=username)}), 401
    user = user[0]
    return video.to_marti_json(user), 200


@video_marti_api.route('/Marti/api/video/<uid>', methods=['DELETE'])
def delete_video(uid):
    video_stream = db.session.execute(db.session.query(VideoStream).filter_by(uid=uid))
    if not video_stream:
        return jsonify({'success': False, 'error': gettext(u'Video stream with uid %(uid)s not found', uid=uid)}), 404

    video_stream = video_stream[0]

    db.session.delete(video_stream)
    db.session.commit()

    return '', 200



from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import Integer, ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column, relationship


@dataclass
class MissionContentMission(db.Model):
    __tablename__ = "mission_content_mission"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    mission_content_id: Mapped[int] = mapped_column(Integer, ForeignKey("mission_content.id"))
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey("missions.name"))

    def serialize(self):
        return {
            "mission_content_id": self.mission_content_id,
            "mission_name": self.mission_name
        }

    def to_json(self):
        return self.serialize()



"""Added publish_time to packages table

Revision ID: 34d5df698b81
Revises: 34dc96ee805b
Create Date: 2024-09-12 17:47:08.855213

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '34d5df698b81'
down_revision = '34dc96ee805b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('device_profiles', schema=None) as batch_op:
        batch_op.alter_column('preference_key',
               existing_type=sa.String(length=255),
               nullable=False)

    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.add_column(sa.Column('publish_time', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.drop_column('publish_time')

    with op.batch_alter_table('device_profiles', schema=None) as batch_op:
        batch_op.alter_column('preference_key',
               existing_type=sa.String(length=255),
               nullable=True)

    # ### end Alembic commands ###



"""Fixed mission_changes and mission_uids relationship

Revision ID: 21fb5a21f356
Revises: 795ebb9262d8
Create Date: 2024-11-08 14:09:41.129561

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '21fb5a21f356'
down_revision = '795ebb9262d8'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_changes', schema=None) as batch_op:
        batch_op.add_column(sa.Column('mission_uid', sa.String(length=255), nullable=True))
        batch_op.create_foreign_key('mission_change_uid', 'mission_uids', ['mission_uid'], ['uid'], ondelete='CASCADE')

    with op.batch_alter_table('mission_uids', schema=None) as batch_op:
        batch_op.drop_constraint('mission_change', type_='foreignkey')
        batch_op.drop_column('mission_change_id')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_uids', schema=None) as batch_op:
        batch_op.add_column(sa.Column('mission_change_id', sa.INTEGER(), nullable=True))
        batch_op.create_foreign_key('mission_change', 'mission_changes', ['mission_change_id'], ['id'])

    with op.batch_alter_table('mission_changes', schema=None) as batch_op:
        batch_op.drop_constraint('mission_change_uid', type_='foreignkey')
        batch_op.drop_column('mission_uid')

    # ### end Alembic commands ###



"""Added data sync tables

Revision ID: 5d06227dea50
Revises: 6af2256c568d
Create Date: 2024-10-09 14:35:53.352679

"""
import uuid

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5d06227dea50'
down_revision = '6af2256c568d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('mission_content',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('keywords', sa.JSON(), nullable=True),
    sa.Column('mime_type', sa.String(length=255), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('submission_time', sa.DateTime(), nullable=True),
    sa.Column('submitter', sa.String(length=255), nullable=True),
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('creator_uid', sa.String(length=255), nullable=True),
    sa.Column('hash', sa.String(length=255), nullable=True),
    sa.Column('size', sa.Integer(), nullable=True),
    sa.Column('expiration', sa.Integer(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uid')
    )
    op.create_table('missions',
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.Column('description', sa.String(length=255), nullable=True),
    sa.Column('chat_room', sa.String(length=255), nullable=True),
    sa.Column('base_layer', sa.String(length=255), nullable=True),
    sa.Column('bbox', sa.String(length=255), nullable=True),
    sa.Column('path', sa.String(length=255), nullable=True),
    sa.Column('classification', sa.String(length=255), nullable=True),
    sa.Column('tool', sa.String(length=255), nullable=True),
    sa.Column('group_name', sa.String(length=255), nullable=True),
    sa.Column('default_role', sa.String(length=255), nullable=True),
    sa.Column('keywords', sa.JSON(), nullable=True),
    sa.Column('creator_uid', sa.String(length=255), nullable=True),
    sa.Column('create_time', sa.DateTime(), nullable=True),
    sa.Column('external_data', sa.JSON(), nullable=True),
    sa.Column('feeds', sa.JSON(), nullable=True),
    sa.Column('map_layers', sa.JSON(), nullable=True),
    sa.Column('invite_only', sa.Boolean(), nullable=True),
    sa.Column('expiration', sa.Integer(), nullable=False),
    sa.Column('guid', sa.String(length=255), nullable=True),
    sa.Column('password_protected', sa.Boolean(), nullable=True),
    sa.Column('password', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('name')
    )
    op.create_table('mission_changes',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('isFederatedChange', sa.Boolean(), nullable=False),
    sa.Column('change_type', sa.String(length=255), nullable=False),
    sa.Column('mission_name', sa.String(length=255), nullable=False),
    sa.Column('timestamp', sa.DateTime(), nullable=False),
    sa.Column('creator_uid', sa.String(length=255), nullable=False),
    sa.Column('server_time', sa.DateTime(), nullable=False),
    sa.ForeignKeyConstraint(['mission_name'], ['missions.name'], name="mission_changes"),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('mission_content_mission',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('mission_content_id', sa.Integer(), nullable=False),
    sa.Column('mission_name', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['mission_content_id'], ['mission_content.id'], name="mission_content"),
    sa.ForeignKeyConstraint(['mission_name'], ['missions.name'], name="mission"),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('mission_invitations',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('mission_name', sa.String(length=255), nullable=False),
    sa.Column('client_uid', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['mission_name'], ['missions.name'], name="mission_invitations"),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('mission_roles',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('clientUid', sa.String(length=255), nullable=False),
    sa.Column('username', sa.String(length=255), nullable=False),
    sa.Column('createTime', sa.DateTime(), nullable=False),
    sa.Column('role_type', sa.String(length=255), nullable=False),
    sa.Column('mission_name', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['mission_name'], ['missions.name'], name="mission_role_names"),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('mission_roles')
    op.drop_table('mission_invitations')
    op.drop_table('mission_content_mission')
    op.drop_table('mission_changes')
    op.drop_table('missions')
    op.drop_table('mission_content')
    # ### end Alembic commands ###



from wtforms import StringField, IntegerField, BooleanField, FloatField, FieldList, FormField
from wtforms.validators import DataRequired, UUID

from opentakserver.forms.point_form import PointForm
from opentakserver.forms.zmist_form import ZmistForm


class CasEvacForm(PointForm):
    uid = StringField(validators=[DataRequired(), UUID()])
    ambulatory = IntegerField()
    casevac = BooleanField()
    child = IntegerField()
    convenience = IntegerField()
    enemy = IntegerField()
    epw = IntegerField()
    equipment_detail = StringField()
    equipment_none = BooleanField()
    equipment_other = BooleanField()
    extraction_equipment = BooleanField()
    freq = FloatField()
    friendlies = StringField()
    hlz_marking = IntegerField()
    hlz_remarks = StringField()
    hoist = BooleanField()
    litter = IntegerField()
    marked_by = StringField()
    medline_remarks = StringField()
    nonus_civilian = IntegerField()
    nonus_military = IntegerField()
    obstacles = StringField()
    priority = IntegerField()
    routine = IntegerField()
    security = IntegerField()
    terrain_loose = BooleanField()
    terrain_other = BooleanField()
    terrain_other_detail = BooleanField()
    terrain_detail = StringField()
    terrain_none = BooleanField()
    terrain_rough = BooleanField()
    terrain_slope = BooleanField()
    terrain_slope_dir = StringField()
    title = StringField(validators=[DataRequired()])
    urgent = IntegerField()
    urgent_surgical = IntegerField()
    us_civilian = IntegerField()
    us_military = IntegerField()
    ventilator = BooleanField()
    winds_are_from = StringField()
    zone_prot_selection = IntegerField()
    zmist = FieldList(FormField(ZmistForm), min_entries=0)



from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, Boolean
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class Marker(db.Model):
    __tablename__ = "markers"

    # type = a-[a-z]-[A-Z]
    # how = h-g-i-g-o

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    uid: Mapped[str] = mapped_column(String(255), unique=True)
    affiliation: Mapped[str] = mapped_column(String(255), nullable=True)
    battle_dimension: Mapped[str] = mapped_column(String(255), nullable=True)
    point_id: Mapped[int] = mapped_column(Integer, ForeignKey("points.id", ondelete="CASCADE"))
    callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    readiness: Mapped[bool] = mapped_column(Boolean, nullable=True)
    argb: Mapped[int] = mapped_column(Integer, nullable=True)
    color_hex: Mapped[str] = mapped_column(String(255), nullable=True)
    iconset_path: Mapped[str] = mapped_column(String(255), nullable=True)
    parent_callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    production_time: Mapped[str] = mapped_column(String(255), nullable=True)
    relation: Mapped[str] = mapped_column(String(255), nullable=True)
    relation_type: Mapped[str] = mapped_column(String(255), nullable=True)
    location_source: Mapped[str] = mapped_column(String(255), nullable=True)
    icon_id: Mapped[int] = mapped_column(Integer, ForeignKey("icons.id"), nullable=True)

    # Will either be the uid attribute in the <Link> tag or
    # if there's no <Link> tag it's assumed that the sender is the parent
    parent_uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True)
    remarks: Mapped[str] = mapped_column(String(255), nullable=True)
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id", ondelete="CASCADE"), nullable=True)
    mil_std_2525c: Mapped[str] = mapped_column(String(255), nullable=True)
    point = relationship("Point", cascade="all, delete", back_populates="marker")
    cot = relationship("CoT", back_populates="marker")
    eud = relationship("EUD", back_populates="markers")
    icon = relationship("Icon", back_populates="markers")

    def color_to_hex(self):
        if self.argb:
            return format(int(self.argb) & 0xFFFFFFFF, '08X')

    def serialize(self):
        return {
            'uid': self.uid,
            'affiliation': self.affiliation,
            'battle_dimension': self.battle_dimension,
            'callsign': self.callsign,
            'readiness': self.readiness,
            'argb': self.argb,
            'color_hex': self.color_hex,
            'iconset_path': self.iconset_path,
            'parent_callsign': self.parent_callsign,
            'relation': self.relation,
            'relation_type': self.relation_type,
            'production_time': self.production_time,
            'location_source': self.location_source,
            'mil_std_2525c': self.mil_std_2525c,
        }

    def to_json(self):
        return {
            'uid': self.uid,
            'affiliation': self.affiliation,
            'battle_dimension': self.battle_dimension,
            'callsign': self.callsign,
            'readiness': self.readiness,
            'argb': self.argb,
            'color_hex': self.color_hex,
            'iconset_path': self.iconset_path,
            'parent_callsign': self.parent_callsign,
            'relation': self.relation,
            'relation_type': self.relation_type,
            'production_time': self.production_time,
            'location_source': self.location_source,
            'icon': self.icon.to_json() if self.icon else None,
            'point': self.point.to_json() if self.point else None,
            'mil_std_2525c': self.mil_std_2525c,
            'type': self.cot.type if self.cot else None,
            'how': self.cot.how if self.cot else None,
            'start': iso8601_string_from_datetime(self.cot.start) if self.cot else None,
            'stale': iso8601_string_from_datetime(self.cot.stale) if self.cot else None,
        }



"""Added fields for ATAK plugin versions

Revision ID: 8d6f811f4fd7
Revises: ab719ee9b322
Create Date: 2025-11-06 15:26:31.710563

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '8d6f811f4fd7'
down_revision = 'ab719ee9b322'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table('packages', schema=None) as batch_op:
            batch_op.add_column(sa.Column('atak_version', sa.String(length=255), nullable=True))
    except BaseException as e:
        print(e)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.drop_column('atak_version')

    # ### end Alembic commands ###



from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, BooleanField, FloatField
from wtforms.validators import DataRequired, UUID


class ZmistForm(FlaskForm):
    i = StringField()
    m = StringField()
    s = StringField()
    t = StringField()
    title = StringField()
    z = StringField()



import datetime
from xml.etree.ElementTree import Element, fromstring, tostring

from flask import Blueprint, request, jsonify
from flask_babel import gettext

from opentakserver.functions import datetime_from_iso8601_string
from opentakserver.extensions import db, logger
from opentakserver.models.CoT import CoT

cot_marti_api = Blueprint('cot_api', __name__)

"""
Right now OpenTAKServer only uses a few of these for Data Sync. The rest were added as place holders 
based on the API docs until I find an example of them actually being used by a TAK client 
"""


@cot_marti_api.route('/Marti/api/cot')
def get_cots():
    return '', 200


@cot_marti_api.route('/Marti/api/cot/xml/<uid>')
def get_cot(uid):
    logger.debug(request.headers)
    logger.debug(request.args)

    cot = db.session.execute(db.session.query(CoT).filter_by(uid=uid)).first()
    if not cot:
        return jsonify({'success': False, 'error': gettext(u"No CoT found for UID %(uid)s", uid=uid)}), 404

    return cot[0].xml


@cot_marti_api.route('/Marti/api/cot/xml/<uid>/all')
def get_all_cot(uid):
    logger.debug(request.headers)
    logger.debug(request.args)

    sec_ago = request.args.get('secago')
    start = request.args.get('start')
    end = request.args.get('end')

    query = db.session.query(CoT).filter_by(uid=uid)
    if sec_ago:
        try:
            query = query.filter(CoT.start >= datetime.timedelta(seconds=int(sec_ago)))
        except ValueError:
            return jsonify({'success': False, 'error': gettext(u'Invalid secago value: %(sec_ago)s', sec_ago=sec_ago)}), 400
    if start:
        query = query.filter(CoT.start >= datetime_from_iso8601_string(start))
    if end:
        query = query.filter(CoT.stale <= datetime_from_iso8601_string(end))

    cots = db.session.execute(query)

    events = Element("events")
    for cot in cots:
        events.append(fromstring(cot[0].xml))

    return tostring(events).decode('utf-8'), 200


@cot_marti_api.route('/Marti/api/cot/sa')
def get_cot_by_time_and_bbox():
    logger.debug(request.headers)
    logger.debug(request.args)

    start = request.args.get('start')
    end = request.args.get('end')
    left = request.args.get('left')
    bottom = request.args.get('bottom')
    right = request.args.get('right')
    top = request.args.get('top')

    return '', 200



import bleach
from flask import Blueprint, current_app as app, request, jsonify

citrap_api_blueprint = Blueprint("citrap_api_blueprint", __name__)


@citrap_api_blueprint.route('/Marti/api/missions/citrap/subscription', methods=['PUT'])
def citrap_subscription():
    uid = bleach.clean(request.args.get('uid'))
    response = {
        'version': 3, 'type': 'com.bbn.marti.sync.model.MissionSubscription',
        'data': {

        }
    }
    return '', 201


@citrap_api_blueprint.route('/Marti/api/citrap')
def citrap():
    return jsonify([])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: atak.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\natak.proto\x12\nmeshtastic\"\xe6\x01\n\tTAKPacket\x12\x15\n\ris_compressed\x18\x01 \x01(\x08\x12$\n\x07\x63ontact\x18\x02 \x01(\x0b\x32\x13.meshtastic.Contact\x12 \n\x05group\x18\x03 \x01(\x0b\x32\x11.meshtastic.Group\x12\"\n\x06status\x18\x04 \x01(\x0b\x32\x12.meshtastic.Status\x12\x1e\n\x03pli\x18\x05 \x01(\x0b\x32\x0f.meshtastic.PLIH\x00\x12#\n\x04\x63hat\x18\x06 \x01(\x0b\x32\x13.meshtastic.GeoChatH\x00\x42\x11\n\x0fpayload_variant\"2\n\x07GeoChat\x12\x0f\n\x07message\x18\x01 \x01(\x0c\x12\x0f\n\x02to\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x42\x05\n\x03_to\"M\n\x05Group\x12$\n\x04role\x18\x01 \x01(\x0e\x32\x16.meshtastic.MemberRole\x12\x1e\n\x04team\x18\x02 \x01(\x0e\x32\x10.meshtastic.Team\"\x19\n\x06Status\x12\x0f\n\x07\x62\x61ttery\x18\x01 \x01(\r\"4\n\x07\x43ontact\x12\x10\n\x08\x63\x61llsign\x18\x01 \x01(\x0c\x12\x17\n\x0f\x64\x65vice_callsign\x18\x02 \x01(\x0c\"_\n\x03PLI\x12\x12\n\nlatitude_i\x18\x01 \x01(\x0f\x12\x13\n\x0blongitude_i\x18\x02 \x01(\x0f\x12\x10\n\x08\x61ltitude\x18\x03 \x01(\x05\x12\r\n\x05speed\x18\x04 \x01(\r\x12\x0e\n\x06\x63ourse\x18\x05 \x01(\r*\xc0\x01\n\x04Team\x12\x14\n\x10Unspecifed_Color\x10\x00\x12\t\n\x05White\x10\x01\x12\n\n\x06Yellow\x10\x02\x12\n\n\x06Orange\x10\x03\x12\x0b\n\x07Magenta\x10\x04\x12\x07\n\x03Red\x10\x05\x12\n\n\x06Maroon\x10\x06\x12\n\n\x06Purple\x10\x07\x12\r\n\tDark_Blue\x10\x08\x12\x08\n\x04\x42lue\x10\t\x12\x08\n\x04\x43yan\x10\n\x12\x08\n\x04Teal\x10\x0b\x12\t\n\x05Green\x10\x0c\x12\x0e\n\nDark_Green\x10\r\x12\t\n\x05\x42rown\x10\x0e*\x7f\n\nMemberRole\x12\x0e\n\nUnspecifed\x10\x00\x12\x0e\n\nTeamMember\x10\x01\x12\x0c\n\x08TeamLead\x10\x02\x12\x06\n\x02HQ\x10\x03\x12\n\n\x06Sniper\x10\x04\x12\t\n\x05Medic\x10\x05\x12\x13\n\x0f\x46orwardObserver\x10\x06\x12\x07\n\x03RTO\x10\x07\x12\x06\n\x02K9\x10\x08\x42_\n\x13\x63om.geeksville.meshB\nATAKProtosZ\"github.com/meshtastic/go/generated\xaa\x02\x14Meshtastic.Protobufs\xba\x02\x00\x62\x06proto3')

_TEAM = DESCRIPTOR.enum_types_by_name['Team']
Team = enum_type_wrapper.EnumTypeWrapper(_TEAM)
_MEMBERROLE = DESCRIPTOR.enum_types_by_name['MemberRole']
MemberRole = enum_type_wrapper.EnumTypeWrapper(_MEMBERROLE)
Unspecifed_Color = 0
White = 1
Yellow = 2
Orange = 3
Magenta = 4
Red = 5
Maroon = 6
Purple = 7
Dark_Blue = 8
Blue = 9
Cyan = 10
Teal = 11
Green = 12
Dark_Green = 13
Brown = 14
Unspecifed = 0
TeamMember = 1
TeamLead = 2
HQ = 3
Sniper = 4
Medic = 5
ForwardObserver = 6
RTO = 7
K9 = 8


_TAKPACKET = DESCRIPTOR.message_types_by_name['TAKPacket']
_GEOCHAT = DESCRIPTOR.message_types_by_name['GeoChat']
_GROUP = DESCRIPTOR.message_types_by_name['Group']
_STATUS = DESCRIPTOR.message_types_by_name['Status']
_CONTACT = DESCRIPTOR.message_types_by_name['Contact']
_PLI = DESCRIPTOR.message_types_by_name['PLI']
TAKPacket = _reflection.GeneratedProtocolMessageType('TAKPacket', (_message.Message,), {
  'DESCRIPTOR' : _TAKPACKET,
  '__module__' : 'atak_pb2'
  # @@protoc_insertion_point(class_scope:meshtastic.TAKPacket)
  })
_sym_db.RegisterMessage(TAKPacket)

GeoChat = _reflection.GeneratedProtocolMessageType('GeoChat', (_message.Message,), {
  'DESCRIPTOR' : _GEOCHAT,
  '__module__' : 'atak_pb2'
  # @@protoc_insertion_point(class_scope:meshtastic.GeoChat)
  })
_sym_db.RegisterMessage(GeoChat)

Group = _reflection.GeneratedProtocolMessageType('Group', (_message.Message,), {
  'DESCRIPTOR' : _GROUP,
  '__module__' : 'atak_pb2'
  # @@protoc_insertion_point(class_scope:meshtastic.Group)
  })
_sym_db.RegisterMessage(Group)

Status = _reflection.GeneratedProtocolMessageType('Status', (_message.Message,), {
  'DESCRIPTOR' : _STATUS,
  '__module__' : 'atak_pb2'
  # @@protoc_insertion_point(class_scope:meshtastic.Status)
  })
_sym_db.RegisterMessage(Status)

Contact = _reflection.GeneratedProtocolMessageType('Contact', (_message.Message,), {
  'DESCRIPTOR' : _CONTACT,
  '__module__' : 'atak_pb2'
  # @@protoc_insertion_point(class_scope:meshtastic.Contact)
  })
_sym_db.RegisterMessage(Contact)

PLI = _reflection.GeneratedProtocolMessageType('PLI', (_message.Message,), {
  'DESCRIPTOR' : _PLI,
  '__module__' : 'atak_pb2'
  # @@protoc_insertion_point(class_scope:meshtastic.PLI)
  })
_sym_db.RegisterMessage(PLI)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\023com.geeksville.meshB\nATAKProtosZ\"github.com/meshtastic/go/generated\252\002\024Meshtastic.Protobufs\272\002\000'
  _TEAM._serialized_start=569
  _TEAM._serialized_end=761
  _MEMBERROLE._serialized_start=763
  _MEMBERROLE._serialized_end=890
  _TAKPACKET._serialized_start=27
  _TAKPACKET._serialized_end=257
  _GEOCHAT._serialized_start=259
  _GEOCHAT._serialized_end=309
  _GROUP._serialized_start=311
  _GROUP._serialized_end=388
  _STATUS._serialized_start=390
  _STATUS._serialized_end=415
  _CONTACT._serialized_start=417
  _CONTACT._serialized_end=469
  _PLI._serialized_start=471
  _PLI._serialized_end=566
# @@protoc_insertion_point(module_scope)



from datetime import datetime, timezone

from sqlalchemy import String, Boolean, DateTime, Integer
from sqlalchemy.orm import Mapped, mapped_column

from opentakserver.extensions import db
from opentakserver.forms.device_profile_form import DeviceProfileForm
from opentakserver.functions import iso8601_string_from_datetime


class DeviceProfiles(db.Model):
    __tablename__ = "device_profiles"

    preference_key: Mapped[str] = mapped_column(String(255), primary_key=True, nullable=False)
    preference_value: Mapped[str] = mapped_column(String(255), nullable=True)
    value_class: Mapped[str] = mapped_column(String(255), nullable=True)
    enrollment: Mapped[bool] = mapped_column(Boolean, default=True, nullable=True)
    connection: Mapped[bool] = mapped_column(Boolean, default=False, nullable=True)
    tool: Mapped[str] = mapped_column(String(255), nullable=True)
    active: Mapped[bool] = mapped_column(Boolean, default=True)
    publish_time: Mapped[datetime] = mapped_column(DateTime)

    def from_wtf(self, form: DeviceProfileForm):
        self.preference_key = form.preference_key.data
        self.preference_value = form.preference_value.data
        self.value_class = f"class java.lang.{form.value_class.data.split('.')[-1]}"
        self.enrollment = form.enrollment.data
        self.connection = form.connection.data
        self.tool = form.tool.data
        self.active = form.active.data
        self.publish_time = datetime.now(timezone.utc)

    def serialize(self):
        return {
            'preference_key': self.preference_key,
            'preference_value': self.preference_value,
            'value_class': self.value_class,
            'enrollment': self.enrollment,
            'connection': self.connection,
            'tool': self.tool,
            'active': self.active,
            'publish_time': self.publish_time
        }

    def to_json(self):
        return_value = self.serialize()
        return_value['publish_time'] = iso8601_string_from_datetime(self.publish_time)
        return return_value



"""Added device_profile and package tables

Revision ID: 414b87c32cc2
Revises: d91957bb59a0
Create Date: 2024-09-05 19:17:20.979466

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '414b87c32cc2'
down_revision = 'd91957bb59a0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('device_profiles',
    sa.Column('preference_key', sa.String(length=255), nullable=False),
    sa.Column('preference_value', sa.String(length=255), nullable=False),
    sa.Column('value_class', sa.String(length=255), nullable=False),
    sa.Column('enrollment', sa.Boolean(), nullable=True),
    sa.Column('connection', sa.Boolean(), nullable=True),
    sa.Column('tool', sa.String(length=255), nullable=True),
    sa.Column('active', sa.Boolean(), nullable=False),
    sa.Column('publish_time', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False),
    sa.PrimaryKeyConstraint('preference_key')
    )
    op.create_table('packages',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('platform', sa.String(length=255), nullable=False),
    sa.Column('plugin_type', sa.String(length=255), nullable=False),
    sa.Column('package_name', sa.String(length=255), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.Column('file_name', sa.String(length=255), nullable=False),
    sa.Column('version', sa.String(length=255), nullable=False),
    sa.Column('revision_code', sa.Integer(), nullable=True),
    sa.Column('description', sa.Integer(), nullable=True),
    sa.Column('apk_hash', sa.Integer(), nullable=True),
    sa.Column('os_requirement', sa.Integer(), nullable=True),
    sa.Column('tak_prereq', sa.Integer(), nullable=True),
    sa.Column('file_size', sa.Integer(), nullable=False),
    sa.Column('icon', sa.LargeBinary(), nullable=True),
    sa.Column('icon_filename', sa.String(length=255), nullable=True),
    sa.Column('install_on_enrollment', sa.Boolean(), nullable=True),
    sa.Column('install_on_connection', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('package_name')
    )
    with op.batch_alter_table('data_packages', schema=None) as batch_op:
        batch_op.add_column(sa.Column('install_on_enrollment', sa.Boolean(), nullable=True))
        batch_op.add_column(sa.Column('install_on_connection', sa.Boolean(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('data_packages', schema=None) as batch_op:
        batch_op.drop_column('install_on_connection')
        batch_op.drop_column('install_on_enrollment')

    op.drop_table('packages')
    op.drop_table('device_profiles')
    # ### end Alembic commands ###



from datetime import datetime

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


class Alert(db.Model):
    __tablename__ = 'alerts'

    # type = ^b-a- (b-a-o-tbl, b-a-o-can)
    # how = ^m-g | ^h-e

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    uid: Mapped[str] = mapped_column(String(255))
    sender_uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"))
    start_time: Mapped[datetime] = mapped_column(DateTime)
    cancel_time: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    alert_type: Mapped[str] = mapped_column(String(255))
    point_id: Mapped[int] = mapped_column(Integer, ForeignKey("points.id", ondelete="CASCADE"), nullable=True)
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id", ondelete="CASCADE"), nullable=True)
    cot = relationship("CoT", back_populates="alert")
    point = relationship("Point", back_populates="alert", foreign_keys=[point_id])
    eud = relationship("EUD", back_populates="alert")

    def serialize(self):
        return {
            'uid': self.uid,
            'sender_uid': self.sender_uid,
            'start_time': self.start_time,
            'cancel_time': self.cancel_time,
            'alert_type': self.alert_type,
            'point_id': self.point_id,
            'cot_id': self.cot_id
        }

    def to_json(self):
        return {
            'uid': self.uid,
            'sender_uid': self.sender_uid,
            'start_time': iso8601_string_from_datetime(self.start_time),
            'cancel_time': iso8601_string_from_datetime(self.cancel_time) if self.cancel_time else None,
            'alert_type': self.alert_type,
            'point': self.point.to_json() if self.point else None,
            'callsign': self.eud.callsign if self.eud else None,
        }



"""Added mission_logs table

Revision ID: 14e268184cc5
Revises: d4cc3d4afdb9
Create Date: 2024-10-15 20:28:12.426250

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '14e268184cc5'
down_revision = 'd4cc3d4afdb9'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('mission_logs',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('content', sa.String(length=255), nullable=False),
    sa.Column('creator_uid', sa.String(length=255), nullable=False),
    sa.Column('entry_uid', sa.String(length=255), nullable=False),
    sa.Column('mission_name', sa.String(length=255), nullable=False),
    sa.Column('server_time', sa.DateTime(), nullable=False),
    sa.Column('dtg', sa.DateTime(), nullable=False),
    sa.Column('created', sa.DateTime(), nullable=False),
    sa.Column('content_hash', sa.String(length=255), nullable=True),
    sa.Column('keywords', sa.JSON(), nullable=False),
    sa.ForeignKeyConstraint(['mission_name'], ['missions.name'], name="mission_logs"),
    sa.PrimaryKeyConstraint('id')
    )
    with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
        batch_op.create_foreign_key('euds', 'euds', ['client_uid'], ['uid'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
        batch_op.drop_constraint('euds', type_='foreignkey')

    op.drop_table('mission_logs')
    # ### end Alembic commands ###



from apscheduler.jobstores.base import ConflictingIdError
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore


class SQLJobStore(SQLAlchemyJobStore):
    def add_job(self, job):
        try:
            super().add_job(job)
        except ConflictingIdError:
            pass



from flask import request, Blueprint, jsonify, current_app as app
from opentakserver.blueprints.ots_api.api import api_blueprint
from opentakserver.blueprints.ots_api.casevac_api import casevac_api_blueprint
from opentakserver.blueprints.ots_api.data_package_api import data_package_api
from opentakserver.blueprints.ots_api.device_profile_api import device_profile_api_blueprint
from opentakserver.blueprints.ots_api.marker_api import marker_api_blueprint
from opentakserver.blueprints.ots_api.mediamtx_api import mediamtx_api_blueprint
from opentakserver.blueprints.ots_api.meshtastic_api import meshtastic_api_blueprint
from opentakserver.blueprints.ots_api.package_api import packages_blueprint
from opentakserver.blueprints.ots_api.scheduler_api import scheduler_api_blueprint
from opentakserver.blueprints.ots_api.token_api import token_api_blueprint
from opentakserver.blueprints.ots_api.user_api import user_api_blueprint
from opentakserver.blueprints.ots_api.video_api import video_api_blueprint
from opentakserver.blueprints.ots_api.mission_api import data_sync_api
from opentakserver.blueprints.ots_api.group_api import group_api
from opentakserver.blueprints.ots_api.eud_stats_api import eud_stats_blueprint
from opentakserver.blueprints.ots_api.plugin_api import plugin_blueprint
from .ldap_api import ldap_blueprint
from .tak_gov_link_api import tak_gov_link_blueprint
from .language_api import language_api

ots_api = Blueprint("ots_api", __name__)
ots_api.register_blueprint(api_blueprint)
ots_api.register_blueprint(casevac_api_blueprint)
ots_api.register_blueprint(data_package_api)
ots_api.register_blueprint(device_profile_api_blueprint)
ots_api.register_blueprint(marker_api_blueprint)
ots_api.register_blueprint(mediamtx_api_blueprint)
ots_api.register_blueprint(meshtastic_api_blueprint)
ots_api.register_blueprint(packages_blueprint)
ots_api.register_blueprint(scheduler_api_blueprint)
ots_api.register_blueprint(user_api_blueprint)
ots_api.register_blueprint(video_api_blueprint)
ots_api.register_blueprint(data_sync_api)
ots_api.register_blueprint(group_api)
ots_api.register_blueprint(eud_stats_blueprint)
ots_api.register_blueprint(plugin_blueprint)
ots_api.register_blueprint(token_api_blueprint)
ots_api.register_blueprint(ldap_blueprint)
ots_api.register_blueprint(tak_gov_link_blueprint)
ots_api.register_blueprint(language_api)


import traceback
from datetime import timedelta, timezone
import json
from uuid import UUID, uuid4

import bleach
import pika
from flask import Blueprint, request, jsonify, current_app as app
from flask_babel import gettext
from flask_security import auth_required, current_user
from sqlalchemy import insert, update
from sqlalchemy.exc import IntegrityError
import xml.etree.ElementTree as ET

from opentakserver.blueprints.ots_api.api import search, paginate, route_cot
from opentakserver.extensions import db, logger, socketio
from opentakserver.functions import *
from opentakserver.models.CoT import CoT
from opentakserver.models.Marker import Marker
from opentakserver.models.Point import Point

marker_api_blueprint = Blueprint('marker_api_blueprint', __name__)


@marker_api_blueprint.route('/api/markers', methods=['GET'])
@auth_required()
def get_markers():
    query = db.session.query(Marker)
    query = search(query, Marker, 'uid')
    query = search(query, Marker, 'affiliation')
    query = search(query, Marker, 'callsign')

    return paginate(query)


@marker_api_blueprint.route('/api/markers', methods=['POST'])
@auth_required()
def add_marker():
    marker = Marker()
    point = Point()

    try:
        if 'latitude' not in request.json.keys() or 'longitude' not in request.json.keys():
            return jsonify({'success': False, 'error': gettext(u'Please provide a latitude and longitude')}), 400
        elif float(request.json['latitude']) < -90 or float(request.json['latitude']) > 90:
            return jsonify({'success': False, 'error': gettext(u"Invalid latitude: %(latitude)s", latitude=request.json['latitude'])}), 400
        elif float(request.json['longitude']) < -180 or float(request.json['longitude']) > 180:
            return jsonify({'success': False, 'error': gettext(u"Invalid longitude: %(longitude)s", longitude=request.json['longitude'])}), 400
    except BaseException as e:
        logger.error(f"Failed to parse lat/lon: {e}")
        return jsonify({'success': False, 'error': gettext(u"Failed to parse lat/lon: %(e)s", e=str(e))}), 400

    if 'uid' not in request.json.keys():
        return jsonify({'success': False, 'error': gettext(u'Please provide a UID')}), 400
    elif 'name' not in request.json.keys():
        return jsonify({'success': False, 'error': gettext(u'Please provide a name')}), 400

    try:
        UUID(request.json['uid'], version=4)
        marker.uid = request.json['uid']
    except ValueError:
        return jsonify({'success': False, 'error': gettext(u"Invalid UID. UIDs need to be in UUID4 format")}), 400

    cot_type = request.json['type'] if 'type' in request.json.keys() else 'a-u-G'

    marker.affiliation = get_affiliation(cot_type)
    marker.battle_dimension = get_battle_dimension(cot_type)
    marker.mil_std_2525c = cot_type_to_2525c(cot_type)

    if not marker.affiliation or not marker.battle_dimension:
        return jsonify({'success': False, 'error': gettext(u"Invalid type: %(cot_type)s", cot_type=cot_type)}), 400

    if 'name' in request.json.keys():
        marker.callsign = bleach.clean(request.json['name'])

    try:
        point.uid = str(uuid4())
        point.device_uid = None
        point.location_source = bleach.clean(
            request.json['location_source']) if 'location_source' in request.json.keys() else ""
        point.latitude = float(request.json['latitude'])
        point.longitude = float(request.json['longitude'])
        point.course = float(request.json['course']) if 'course' in request.json.keys() else 0
        point.azimuth = float(request.json['azimuth']) if 'azimuth' in request.json.keys() else 0
        point.speed = float(request.json['speed']) if 'speed' in request.json.keys() else 0
        point.battery = float(request.json['battery']) if 'battery' in request.json.keys() else None
        point.fov = float(request.json['fov']) if 'fov' in request.json.keys() else None
        point.ce = float(request.json['ce']) if 'ce' in request.json.keys() else 9999999.0
        point.hae = float(request.json['hae']) if 'hae' in request.json.keys() else 9999999.0
        point.le = float(request.json['le']) if 'le' in request.json.keys() else 9999999.0
        point.timestamp = datetime.now(timezone.utc)

        with app.app_context():
            event = ET.Element("event")
            event.set("type", cot_type)
            event.set("version", "2.0")
            event.set("how", "m-g")
            event.set("uid", marker.uid)
            event.set("time", iso8601_string_from_datetime(point.timestamp))
            event.set("start", iso8601_string_from_datetime(point.timestamp))
            event.set("stale", iso8601_string_from_datetime(point.timestamp + timedelta(days=1)))

            cot_point = ET.SubElement(event, "point")
            cot_point.set("ce", str(point.ce))
            cot_point.set("hae", str(point.hae))
            cot_point.set("le", str(point.le))
            cot_point.set("lat", str(point.latitude))
            cot_point.set("lon", str(point.longitude))

            detail = ET.SubElement(event, "detail")
            detail.set("uid", marker.uid)

            contact = ET.SubElement(detail, "contact")
            contact.set("callsign", marker.callsign)

            track = ET.SubElement(detail, "track")
            track.set("course", str(point.course))
            track.set("speed", str(point.speed))

            if point.azimuth or point.fov:
                sensor = ET.SubElement(detail, "sensor")
                sensor.set("azimuth", str(point.azimuth))
                sensor.set("fov", str(point.fov))
                sensor.set("fovBlue", "1.0")
                sensor.set("fovGreen", "1.0")
                sensor.set("fovRed", "1.0")
                sensor.set("range", "100.0")

            rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
            rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
            rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
            channel = rabbit_connection.channel()
            channel.basic_publish(exchange='cot', routing_key='', body=json.dumps(
                {'cot': ET.tostring(event).decode('utf-8'), 'uid': app.config['OTS_NODE_ID']}),
                                  properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))
            route_cot(ET.tostring(event).decode('utf-8'), current_user)
            channel.close()
            rabbit_connection.close()

            cot = db.session.execute(insert(CoT).values(
                how="m-g", type=cot_type, timestamp=point.timestamp, xml=ET.tostring(event), start=point.timestamp,
                stale=point.timestamp + timedelta(days=1), sender_callsign=current_user.username
            ))
            db.session.commit()
            marker.cot_id = cot.inserted_primary_key[0]

            p = db.session.execute(insert(Point).values(
                uid=point.uid, device_uid=point.device_uid, ce=point.ce, hae=point.hae, le=point.le,
                latitude=point.latitude,
                longitude=point.longitude, timestamp=point.timestamp, location_source=point.location_source,
                course=point.course, speed=point.speed, battery=point.battery, fov=point.fov, azimuth=point.azimuth,
                cot_id=cot.inserted_primary_key[0]
            ))
            # Commit here to get the point's auto increment primary key
            db.session.commit()

            marker.point_id = p.inserted_primary_key[0]
            try:
                db.session.add(marker)
                db.session.commit()
            except IntegrityError:
                db.session.rollback()
                db.session.execute(update(Marker).where(Marker.uid == marker.uid).values(
                    point_id=marker.point_id,
                    cot_id=marker.cot_id,
                    **marker.serialize()
                ))
                db.session.commit()

            # Get the marker with its associated point and cot data
            marker = db.session.execute(db.session.query(Marker).filter_by(uid=marker.uid)).first()[0]
            socketio.emit('marker', marker.to_json(), namespace='/socket.io')

            return jsonify({'success': True})

    except BaseException as e:
        logger.error(f"Failed to parse data: {e}")
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': gettext(u"Failed to parse data: %(e)s", e=str(e))}), 400


@marker_api_blueprint.route('/api/markers', methods=['DELETE'])
@auth_required()
def delete_marker():
    uid = request.args.get('uid')
    if not uid:
        return jsonify({'success': False, 'error': gettext(u'Please provide the UID of the marker to delete')}), 400

    with app.app_context():
        query = db.session.query(Marker)
        query = search(query, Marker, 'uid')
        marker = db.session.execute(query).first()
        if not marker:
            return jsonify({'success': False, 'error': gettext(u'Unknown UID')}), 404

        marker = marker[0]
        now = datetime.now(timezone.utc)
        event = ET.Element('event', {'how': 'h-g-i-g-o', 'type': 't-x-d-d', 'version': '2.0',
                                     'uid': marker.uid, 'start': iso8601_string_from_datetime(now),
                                     'time': iso8601_string_from_datetime(now),
                                     'stale': iso8601_string_from_datetime(now + timedelta(minutes=10))})
        ET.SubElement(event, 'point', {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0',
                                       'lon': '0'})
        detail = ET.SubElement(event, 'detail')
        ET.SubElement(detail, 'link', {'relation': 'p-p', 'uid': marker.uid, 'type': marker.cot.type})
        ET.SubElement(detail, '_flow-tags_', {'TAK-Server-f1a8159ef7804f7a8a32d8efc4b773d0': iso8601_string_from_datetime(now)})

        rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
        rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
        rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
        channel = rabbit_connection.channel()
        channel.basic_publish(exchange='cot', routing_key='', body=json.dumps(
            {'cot': ET.tostring(event).decode('utf-8'), 'uid': app.config['OTS_NODE_ID']}),
                              properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))
        route_cot(ET.tostring(event).decode('utf-8'), current_user)
        channel.close()
        rabbit_connection.close()

        db.session.delete(marker)
        db.session.commit()

        return jsonify({'success': True})



import json
import uuid
from dataclasses import dataclass
from urllib.parse import urlparse
from xml.etree.ElementTree import Element, SubElement, tostring

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, Boolean, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship
from flask import current_app as app, request


@dataclass
class VideoStream(db.Model):
    __tablename__ = 'video_streams'

    path: Mapped[str] = mapped_column(String(255), primary_key=True)
    protocol: Mapped[str] = mapped_column(String(255), default='rtsp')
    port: Mapped[int] = mapped_column(Integer, default=8554)
    network_timeout: Mapped[int] = mapped_column(Integer, default=10000)
    uid: Mapped[str] = mapped_column(String(255), nullable=True, default=str(uuid.uuid4()))
    buffer_time: Mapped[int] = mapped_column(Integer, nullable=True)
    rover_port: Mapped[int] = mapped_column(Integer, nullable=True)
    rtsp_reliable: Mapped[int] = mapped_column(Integer, nullable=True, default=1)
    ignore_embedded_klv: Mapped[bool] = mapped_column(Boolean, nullable=True)
    alias: Mapped[str] = mapped_column(String(255), nullable=True)
    preferred_mac_address: Mapped[str] = mapped_column(String(255), nullable=True)
    preferred_interface_address: Mapped[str] = mapped_column(String(255), nullable=True)
    username: Mapped[str] = mapped_column(String(255), ForeignKey("user.username"), nullable=True)
    xml: Mapped[str] = mapped_column(TEXT, nullable=True)
    ready: Mapped[bool] = mapped_column(Boolean, default=False)
    mediamtx_settings: Mapped[str] = mapped_column(TEXT, default="")
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id"), nullable=True)
    cot = relationship("CoT", back_populates="video")
    user = relationship("User", back_populates="video_streams")
    recordings = relationship("VideoRecording", back_populates="video_stream")

    def serialize(self):
        return {
            'network_timeout': self.network_timeout,
            'uid': self.uid,
            'protocol': self.protocol,
            'path': self.path,
            'buffer_time': self.buffer_time,
            'port': self.port,
            'rover_port': self.rover_port,
            'rtsp_reliable': self.rtsp_reliable,
            'ignore_embedded_klv': self.ignore_embedded_klv,
            'alias': self.alias,
            'preferred_mac_address': self.preferred_mac_address,
            'preferred_interface_address': self.preferred_interface_address,
            'username': self.username,
            'ready': self.ready,
        }

    def to_json(self):
        try:
            mediamtx_settings = json.loads(self.mediamtx_settings)
            source = mediamtx_settings['source']
            record = mediamtx_settings['record']
        except json.decoder.JSONDecodeError:
            source = None
            record = False

        with app.app_context():
            url = urlparse(request.url_root)
            protocol = url.scheme
            hostname = url.hostname
            port = url.port
            if not port and protocol == 'https':
                port = 443
            elif not port and protocol == 'http':
                port = 80

            return {
                'network_timeout': self.network_timeout,
                'uid': self.uid,
                'protocol': self.protocol,
                'path': self.path,
                'buffer_time': self.buffer_time,
                'port': self.port,
                'rover_port': self.rover_port,
                'rtsp_reliable': self.rtsp_reliable,
                'ignore_embedded_klv': self.ignore_embedded_klv,
                'alias': self.alias,
                'preferred_mac_address': self.preferred_mac_address,
                'preferred_interface_address': self.preferred_interface_address,
                'username': self.username,
                'ready': self.ready,
                'source': source,
                'record': record,
                'rtsp_link': "rtsp://{}:{}/{}".format(hostname, self.port, self.path),
                'webrtc_link': "{}://{}:{}/webrtc/{}/".format(protocol, hostname, port, self.path),
                'hls_link': "{}://{}:{}/hls/{}/".format(protocol, hostname, port, self.path),
                'thumbnail': f"{protocol}://{hostname}:{port}/api/videos/thumbnail?path={self.path}"
            }

    def to_marti_json(self, user):
        url = urlparse(request.url_root)
        protocol = url.scheme
        hostname = url.hostname
        port = url.port
        if not port and protocol == 'https':
            port = 443
        elif not port and protocol == 'http':
            port = 80

        video_uuid = str(uuid.uuid4())

        return {
            "active": True,
            "alias": self.path,
            "thumbnail": "",
            "classification": "",
            "feeds": [{"uuid": video_uuid,
                      "active": True,
                       "alias": self.path,
                       "url": f"{protocol}://{hostname}:{port}/hls/{self.path}/index.m3u8?jwt={user.get_auth_token()}",
                       "order": 0,
                       "macAddress": self.preferred_mac_address,
                       "roverPort": str(self.rover_port),
                       "ignoreEmbeddedKLV": str(self.ignore_embedded_klv),
                       "source": f"{protocol}://{hostname}:{port}/hls/{self.path}/index.m3u8",
                       "networkTimeout": str(self.network_timeout),
                       "bufferTime": str(self.buffer_time),
                       "rtspReliable": str(self.rtsp_reliable),
                       "thumbnail": "",
                       "classification": "",
                       "latitude": "",
                       "longitude": "",
                       "fov": "",
                       "heading": "",
                       "range": "",
                       "width": 0,
                       "height": 0,
                       "bitrate": 0}],
            "uuid": video_uuid
        }

    def generate_xml(self, hostname):

        feed = Element('feed')
        # Force rtsp to ensure compatibility with ATAK
        SubElement(feed, 'protocol').text = 'rtsp'
        SubElement(feed, 'alias').text = self.alias if self.alias else self.path
        SubElement(feed, 'uid').text = str(self.uid) if self.uid else str(uuid.uuid4())
        SubElement(feed, 'address').text = hostname
        SubElement(feed, 'port').text = str(self.port) if self.port else "8554"
        SubElement(feed, 'roverPort').text = str(self.rover_port)
        SubElement(feed, 'ignoreEmbeddedKLV').text = self.ignore_embedded_klv
        SubElement(feed, 'preferredMacAddress').text = self.preferred_mac_address
        SubElement(feed, 'preferredInterfaceAddress').text = self.preferred_interface_address
        SubElement(feed, 'path').text = self.path if self.path else self.alias
        SubElement(feed, 'buffer').text = str(self.buffer_time) if self.buffer_time else ""
        SubElement(feed, 'timeout').text = str(self.network_timeout) if self.network_timeout else "10000"
        SubElement(feed, 'rtspReliable').text = str(self.rtsp_reliable) if self.rtsp_reliable else "1"

        self.xml = tostring(feed).decode('utf-8')



import datetime
import io
import os
import re
import traceback
import uuid
import zipfile
from urllib.parse import urlparse

from xml.etree.ElementTree import Element, SubElement, tostring

from flask import Blueprint, request, current_app as app, Response
from werkzeug.wsgi import FileWrapper

import opentakserver
from opentakserver.blueprints.marti_api.marti_api import verify_client_cert
from opentakserver.extensions import db, logger, ldap_manager
from opentakserver.models.DeviceProfiles import DeviceProfiles
from opentakserver.models.Packages import Packages
from opentakserver.models.DataPackage import DataPackage

device_profile_marti_api_blueprint = Blueprint('device_profile_marti_api_blueprint', __name__)


def get_ldap_attributes(prefs: Element):
    if app.config.get("OTS_ENABLE_LDAP"):
        cert = verify_client_cert()
        if not cert:
            return

        username = cert.get_subject().commonName
        user_info = ldap_manager.get_user_info_for_username(username)

        for field in user_info:
            if field.lower() == app.config.get("OTS_LDAP_COLOR_ATTRIBUTE").lower():
                team = SubElement(prefs, "entry", {"key": "locationTeam", "class": "class java.lang.String"})
                team.text = user_info[field][0]

            elif field.lower() == app.config.get("OTS_LDAP_ROLE_ATTRIBUTE").lower():
                role = SubElement(prefs, "entry", {"key": "atakRoleType", "class": "class java.lang.String"})
                role.text = user_info[field][0]

            elif field.lower() == app.config.get("OTS_LDAP_CALLSIGN_ATTRIBUTE").lower():
                callsign = SubElement(prefs, "entry", {"key": "locationCallsign", "class": "class java.lang.String"})
                callsign.text = user_info[field][0]

            elif field.lower().startswith(app.config.get("OTS_LDAP_PREFERENCE_ATTRIBUTE_PREFIX").lower()):
                # Remove prefix, case insensitive
                remove_prefix = re.compile(re.escape(app.config.get("OTS_LDAP_PREFERENCE_ATTRIBUTE_PREFIX")), re.IGNORECASE)
                attribute = remove_prefix.sub("", field)
                subelement = SubElement(prefs, "entry", {"key": attribute, "class": "class java.lang.String"})
                subelement.text = user_info[field][0]


def create_profile_zip(enrollment=True, syncSecago=-1):
    # preference.pref
    prefs = Element("preferences")
    pref = SubElement(prefs, "preference", {"version": "1", "name": "com.atakmap.app_preferences"})

    get_ldap_attributes(pref)

    plugins = None
    data_packages = None
    device_profiles = None
    zip_buffer = io.BytesIO()
    zipf = zipfile.ZipFile(zip_buffer, "a", zipfile.ZIP_DEFLATED, False)

    # MANIFEST file
    manifest = Element("MissionPackageManifest", {"version": "2"})
    config = SubElement(manifest, "Configuration")
    SubElement(config, "Parameter", {"name": "uid", "value": str(uuid.uuid4())})
    SubElement(config, "Parameter", {"name": "name", "value": "Device Profile"})
    SubElement(config, "Parameter", {"name": "onReceiveDelete", "value": "true"})

    contents = SubElement(manifest, "Contents")
    SubElement(contents, "Content", {"ignore": "false", "zipEntry": "5c2bfcae3d98c9f4d262172df99ebac5/preference.pref"})
    SubElement(contents, "Content",
               {"ignore": "false", "zipEntry": "5c2bfcae3d98c9f4d262172df99ebac5/truststore-root.p12"})


    enable_update_server = SubElement(pref, "entry", {"key": "appMgmtEnableUpdateServer", "class": "class java.lang.Boolean"})
    enable_update_server.text = "true"

    update_server_address = SubElement(pref, "entry", {"key": "atakUpdateServerUrl", "class": "class java.lang.String"})
    update_server_address.text = f"https://{urlparse(request.url_root).hostname}:{app.config.get('OTS_MARTI_HTTPS_PORT')}/api/packages"

    startup_sync = SubElement(pref, "entry", {"key": "repoStartupSync", "class": "class java.lang.Boolean"})
    startup_sync.text = "true"

    enable_profiles = SubElement(pref, "entry",
                                 {"key": "deviceProfileEnableOnConnect", "class": "class java.lang.Boolean"})
    enable_profiles.text = "true"

    ca_location = SubElement(pref, "entry", {"key": "updateServerCaLocation", "class": "class java.lang.String"})
    ca_location.text = "/storage/emulated/0/atak/cert/truststore-root.p12"

    ca_password = SubElement(pref, "entry", {"key": "updateServerCaPassword", "class": "class java.lang.String"})
    ca_password.text = app.config.get("OTS_CA_PASSWORD")

    enable_channels_host = SubElement(pref, "entry",
                                      {'key': f'prefs_enable_channels_host-{urlparse(request.url_root).hostname}',
                                       'class': 'class java.lang.String'})
    enable_channels_host.text = "true"

    enable_channels = SubElement(pref, "entry", {'key': 'prefs_enable_channels', 'class': 'class java.lang.String'})
    enable_channels.text = "true" if app.config.get("OTS_ENABLE_CHANNELS") else "false"

    if enrollment:
        # Add the maps to the zip
        if app.config.get("OTS_PROFILE_MAP_SOURCES") and enrollment:
            maps_path = os.path.join(os.path.dirname(opentakserver.__file__), "maps")
            for root, dirs, map_files in os.walk(maps_path):
                for map in map_files:
                    zipf.writestr(f"maps/{map}", open(os.path.join(maps_path, map), 'r').read())
                    SubElement(contents, "Content", {"ignore": "false", "zipEntry": f"maps/{map}"})

            device_profiles = db.session.execute(
                db.session.query(DeviceProfiles).filter_by(enrollment=True, active=True)).all()
            plugins = db.session.execute(db.session.query(Packages).filter_by(install_on_enrollment=True)).all()
            data_packages = db.session.execute(db.session.query(DataPackage).filter_by(install_on_enrollment=True)).all()
    elif syncSecago > 0:
        publish_time = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(seconds=syncSecago)

        device_profiles = db.session.execute(db.session.query(DeviceProfiles).filter_by(connection=True, active=True)
                                             .filter(DeviceProfiles.publish_time >= publish_time)).all()
        data_packages = db.session.execute(db.session.query(DataPackage).filter_by(install_on_connection=True)
                                           .filter(DataPackage.submission_time >= publish_time)).all()
        plugins = db.session.execute(db.session.query(Packages).filter_by(install_on_connection=True)
                                     .filter(Packages.publish_time >= publish_time)).all()
    else:
        device_profiles = db.session.execute(
            db.session.query(DeviceProfiles).filter_by(connection=True, active=True)).all()
        data_packages = db.session.execute(db.session.query(DataPackage).filter_by(install_on_connection=True)).all()

    for profile in device_profiles:
        p = SubElement(pref, "entry", {"key": profile[0].preference_key, "class": profile[0].value_class})
        p.text = profile[0].preference_value

    if plugins:
        for plugin in plugins:
            plugin = plugin[0]

            SubElement(contents, "Content",
                       {"ignore": "false", "zipEntry": f"5c2bfcae3d98c9f4d262172df99ebac5/{plugin.file_name}"})
            zipf.write(os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", plugin.file_name),
                       f"5c2bfcae3d98c9f4d262172df99ebac5/{plugin.file_name}")

    if data_packages:
        for data_package in data_packages:
            data_package = data_package[0]
            SubElement(contents, "Content",
                       {"ignore": "false", "zipEntry": f"5c2bfcae3d98c9f4d262172df99ebac5/{data_package.filename}"})
            zipf.write(os.path.join(app.config.get("UPLOAD_FOLDER"), f"{data_package.hash}.zip"),
                       f"5c2bfcae3d98c9f4d262172df99ebac5/{data_package.filename}")

    zipf.writestr("MANIFEST/manifest.xml", tostring(manifest))

    zipf.writestr("5c2bfcae3d98c9f4d262172df99ebac5/preference.pref", tostring(prefs))

    with open(os.path.join(app.config.get("OTS_CA_FOLDER"), 'truststore-root.p12'), 'rb') as truststore:
        zipf.writestr("5c2bfcae3d98c9f4d262172df99ebac5/truststore-root.p12", truststore.read())

    return zip_buffer


# Authentication for /Marti endpoints handled by client cert validation
# EUDs hit this endpoint after a successful certificate enrollment
@device_profile_marti_api_blueprint.route('/Marti/api/tls/profile/enrollment')
def enrollment_profile():
    try:
        profile_zip = create_profile_zip()
        profile_zip.seek(0)
        return Response(FileWrapper(profile_zip), mimetype="application/zip", direct_passthrough=True)
    except BaseException as e:
        logger.error(f"Failed to send enrollment package: {e}")
        logger.debug(traceback.format_exc())
        return '', 500


# EUDs hit this endpoint when the app connects to the server if repoStartupSync is enabled
@device_profile_marti_api_blueprint.route('/api/connection')
@device_profile_marti_api_blueprint.route('/Marti/api/device/profile/connection')
def connection_profile():
    try:
        syncSecago = -1
        if 'syncSecago' in request.args:
            try:
                syncSecago = int(request.args['syncSecago'])
            except ValueError:
                pass
        profile_zip = create_profile_zip(False, syncSecago)
        profile_zip.seek(0)
        return Response(FileWrapper(profile_zip), mimetype="application/zip", direct_passthrough=True)
    except BaseException as e:
        logger.error(f"Failed to send enrollment package: {e}")
        logger.debug(traceback.format_exc())
        return '', 500



from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import String, Integer, Boolean
from sqlalchemy.orm import Mapped, mapped_column


@dataclass
class MeshtasticChannel(db.Model):
    __tablename__ = "meshtastic_channels"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    psk: Mapped[str] = mapped_column(String(255), nullable=True)
    name: Mapped[str] = mapped_column(String(255), nullable=True)
    uplink_enabled: Mapped[bool] = mapped_column(Boolean, default=False)
    downlink_enabled: Mapped[bool] = mapped_column(Boolean, default=False)
    position_precision: Mapped[int] = mapped_column(Integer, default=32)  # LOW = 11, MED = 16, HIGH = 32, DISABLED = 0
    lora_region: Mapped[int] = mapped_column(Integer, default=0)
    lora_hop_limit: Mapped[int] = mapped_column(Integer, default=3)
    lora_tx_enabled: Mapped[bool] = mapped_column(Boolean, default=True)
    lora_tx_power: Mapped[int] = mapped_column(Integer, default=30)
    lora_sx126x_rx_boosted_gain: Mapped[bool] = mapped_column(Boolean, default=True)
    modem_preset: Mapped[int] = mapped_column(Integer, default=0)
    url: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)

    def serialize(self):
        return {
            'psk': self.psk if self.psk else None,
            'name': self.name,
            'uplink_enabled': self.uplink_enabled,
            'downlink_enabled': self.downlink_enabled,
            'position_precision': self.position_precision,
            'lora_region': self.lora_region,
            'lora_hop_limit': self.lora_hop_limit,
            'lora_tx_enabled': self.lora_tx_enabled,
            'lora_tx_power': self.lora_tx_power,
            'lora_sx126x_rx_boosted_gain': self.lora_sx126x_rx_boosted_gain,
            'modem_preset': self.modem_preset,
            'url': self.url
        }

    def to_json(self):
        preset = "LONG_FAST"
        if self.modem_preset == 1:
            preset = "LONG_SLOW"
        elif self.modem_preset == 2:
            preset = "VERY_LONG_SLOW"
        elif self.modem_preset == 3:
            preset = "MEDIUM_SLOW"
        elif self.modem_preset == 4:
            preset = "MEDIUM_FAST"
        elif self.modem_preset == 5:
            preset = "SHORT_SLOW"
        elif self.modem_preset == 6:
            preset = "SHORT_FAST"
        elif self.modem_preset == 7:
            preset = "LONG_MODERATE"
        elif self.modem_preset == 8:
            preset = "SHORT_TURBO"

        region = "UNSET"
        if self.lora_region == 1:
            region = "US"
        elif self.lora_region == 2:
            region = "EU_433"
        elif self.lora_region == 3:
            region = "EU_868"
        elif self.lora_region == 4:
            region = "CN"
        elif self.lora_region == 5:
            region = "JP"
        elif self.lora_region == 6:
            region = "ANZ"
        elif self.lora_region == 7:
            region = "KR"
        elif self.lora_region == 8:
            region = "TW"
        elif self.lora_region == 9:
            region = "RU"
        elif self.lora_region == 10:
            region = "IN"
        elif self.lora_region == 11:
            region = "NZ_865"
        elif self.lora_region == 12:
            region = "TH"
        elif self.lora_region == 13:
            region = "LORA_24"
        elif self.lora_region == 14:
            region = "UA_433"
        elif self.lora_region == 15:
            region = "UA_868"
        elif self.lora_region == 16:
            region = "MY_433"
        elif self.lora_region == 17:
            region = "MY_919"
        elif self.lora_region == 18:
            region = "SG_923"

        return {
            'psk': self.psk if self.psk else None,
            'name': self.name,
            'uplink_enabled': self.uplink_enabled,
            'downlink_enabled': self.downlink_enabled,
            'position_precision': self.position_precision,
            'lora_region': region,
            'lora_hop_limit': self.lora_hop_limit,
            'lora_tx_enabled': self.lora_tx_enabled,
            'lora_tx_power': self.lora_tx_power,
            'lora_sx126x_rx_boosted_gain': self.lora_sx126x_rx_boosted_gain,
            'modem_preset': preset,
            'url': self.url
        }



from flask import Blueprint

from opentakserver.blueprints.marti_api.certificate_enrollment_api import certificate_authority_api_blueprint
from opentakserver.blueprints.marti_api.citrap_api import citrap_api_blueprint
from opentakserver.blueprints.marti_api.cot_marti_api import cot_marti_api
from opentakserver.blueprints.marti_api.data_package_marti_api import data_package_marti_api
from opentakserver.blueprints.marti_api.mission_marti_api import mission_marti_api
from opentakserver.blueprints.marti_api.device_profile_marti_api import device_profile_marti_api_blueprint
from opentakserver.blueprints.marti_api.group_marti_api import group_api
from opentakserver.blueprints.marti_api.marti_api import marti_api
from opentakserver.blueprints.marti_api.contacts_marti_api import contacts_api
from opentakserver.blueprints.marti_api.video_marti_api import video_marti_api

marti_blueprint = Blueprint("marti_blueprint", __name__)

marti_blueprint.register_blueprint(data_package_marti_api)
marti_blueprint.register_blueprint(marti_api)
marti_blueprint.register_blueprint(cot_marti_api)
marti_blueprint.register_blueprint(device_profile_marti_api_blueprint)
marti_blueprint.register_blueprint(citrap_api_blueprint)
marti_blueprint.register_blueprint(certificate_authority_api_blueprint)
marti_blueprint.register_blueprint(group_api)
marti_blueprint.register_blueprint(mission_marti_api)
marti_blueprint.register_blueprint(contacts_api)
marti_blueprint.register_blueprint(video_marti_api)



import base64
import datetime
import json
import traceback
import uuid

import pika
import sqlalchemy.exc
import unishox2
import os

from meshtastic import mqtt_pb2, portnums_pb2, mesh_pb2, protocols, BROADCAST_NUM

from opentakserver.extensions import db, logger
from opentakserver.models.Meshtastic import MeshtasticChannel
from opentakserver.models.Team import Team
from opentakserver.proto import atak_pb2
from google.protobuf.json_format import MessageToJson
from xml.etree.ElementTree import Element, SubElement, tostring

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend

from opentakserver.controllers.rabbitmq_client import RabbitMQClient
from opentakserver.models.EUD import EUD


class MeshtasticController(RabbitMQClient):
    def __init__(self, context):
        super().__init__(context)
        self.node_names = {}
        self.logger.info("Starting Meshtastic controller...")
        self.meshtastic_devices = {}
        self.get_euds()
        self.get_channels()

    def get_euds(self):
        with self.context:
            euds = self.db.session.execute(self.db.session.query(EUD)).scalars()
            for eud in euds:
                meshtastic_id = eud.meshtastic_id
                if not eud.meshtastic_id:
                    eud.meshtastic_id = int.from_bytes(os.urandom(4), 'big')
                    self.db.session.add(eud)
                    self.db.session.commit()
                self.meshtastic_devices[eud.uid] = {'hw_model': eud.device, 'long_name': eud.callsign, 'short_name': '',
                                                    'firmware_version': eud.version, 'last_lat': "0.0",
                                                    'last_lon': "0.0",
                                                    'battery': 0, 'meshtastic_id': '{:x}'.format(eud.meshtastic_id),
                                                    'voltage': 0,
                                                    'uptime': 0, 'last_alt': "9999999.0", 'course': '0.0',
                                                    'speed': '0.0', 'team': 'Cyan', 'role': 'Team Member',
                                                    'uid': eud.uid,
                                                    'macaddr': eud.meshtastic_macaddr}

    def get_channels(self):
        with self.context:
            channels = self.db.session.execute(self.db.session.query(MeshtasticChannel)).scalars()
            downlink_channels = []
            for channel in channels:
                if channel.downlink_enabled:
                    downlink_channels.append(channel.name)
            self.context.app.config.update({"OTS_MESHTASTIC_DOWNLINK_CHANNELS": downlink_channels})

    def on_channel_open(self, channel):
        self.rabbit_channel = channel
        self.rabbit_channel.queue_declare(queue='meshtastic')
        self.rabbit_channel.queue_bind(exchange='amq.topic', queue='meshtastic', routing_key="#")
        self.rabbit_channel.basic_consume(queue='meshtastic', on_message_callback=self.on_message, auto_ack=True)
        self.rabbit_channel.add_on_close_callback(self.on_close)

    def try_decode(self, mp):
        # Get the channel key from the DB
        key_bytes = base64.b64decode("1PG7OiApB1nwvP+rz05pAQ==".encode('ascii'))

        nonce = getattr(mp, "id").to_bytes(8, "little") + getattr(mp, "from").to_bytes(8, "little")
        cipher = Cipher(algorithms.AES(key_bytes), modes.CTR(nonce), backend=default_backend())
        decryptor = cipher.decryptor()
        decrypted_bytes = decryptor.update(getattr(mp, "encrypted")) + decryptor.finalize()

        data = mesh_pb2.Data()
        data.ParseFromString(decrypted_bytes)
        mp.decoded.CopyFrom(data)

    def on_message(self, unused_channel, basic_deliver, properties, body):
        # Don't process outgoing message from TAK EUDs to the Meshtastic Network, only messages from the Meshtastic
        # network to TAK EUDs
        if basic_deliver.routing_key.endswith('outgoing'):
            return

        # Forward this Meshtastic message to other Meshtastic channels which have downlink enabled
        for channel in self.context.app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS"):
            routing_key = "{}.2.e.{}.".format(self.context.app.config.get("OTS_MESHTASTIC_TOPIC"), channel)
            if not basic_deliver.routing_key.startswith(routing_key):
                self.rabbit_channel.basic_publish(exchange='amq.topic', routing_key=routing_key + "outgoing", body=body,
                                                  properties=pika.BasicProperties(expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")))

        se = mqtt_pb2.ServiceEnvelope()
        try:
            se.ParseFromString(body)
            mp = se.packet
        except Exception as e:
            self.logger.error(f"ERROR: parsing service envelope: {str(e)}")
            self.logger.error(f"{body}")
            return

        meshtastic_id = getattr(mp, 'from')
        meshtastic_id = f"{meshtastic_id:08x}"
        to_id = mp.to
        if to_id == BROADCAST_NUM:
            to_id = 'all'
        else:
            to_id = f"{to_id:08x}"

        pn = portnums_pb2.PortNum.Name(mp.decoded.portnum)

        prefix = f"{mp.channel} [{meshtastic_id}->{to_id}] {pn}:"
        if mp.HasField("encrypted") and not mp.HasField("decoded"):
            try:
                self.try_decode(mp)
                pn = portnums_pb2.PortNum.Name(mp.decoded.portnum)
                prefix = f"{mp.channel} [{meshtastic_id}->{to_id}] {pn}:"
            except Exception as e:
                self.logger.warning(f"{prefix} could not be decrypted")
                return

        handler = protocols.get(mp.decoded.portnum)
        if handler is None:
            try:
                if portnums_pb2.PortNum.Name(mp.decoded.portnum) == "ATAK_PLUGIN":
                    tak_packet = atak_pb2.TAKPacket()
                    tak_packet.ParseFromString(mp.decoded.payload)
                    self.protobuf_to_cot(tak_packet, meshtastic_id, to_id, pn, meshtastic_id)
                    self.logger.info(tak_packet)
            except:
                self.logger.error(traceback.format_exc())

            return

        if handler.protobufFactory is None:
            self.logger.debug(f"{prefix} {mp}")
            self.protobuf_to_cot(mp.decoded.payload, meshtastic_id, to_id, pn, meshtastic_id)
            self.logger.info(mp.decoded.payload)
        else:
            try:
                pb = handler.protobufFactory()
                pb.ParseFromString(mp.decoded.payload)
                p = MessageToJson(pb)
                if mp.decoded.portnum == portnums_pb2.PortNum.NODEINFO_APP:
                    self.node_names[getattr(mp, "from")] = pb.short_name
                    prefix = f"{mp.channel} [{meshtastic_id}->{to_id}] {pn}:"
                    self.rabbit_channel.queue_declare(queue=meshtastic_id)
                self.logger.debug(f"{prefix} {p}")
                self.protobuf_to_cot(pb, meshtastic_id, to_id, pn, meshtastic_id)
                self.logger.info(pb)
            except:
                self.logger.error(traceback.format_exc())

    def cot(self, pb, from_id, to_id, portnum, how='m-g', cot_type='a-f-G-U-C', uid=None):
        if not uid and from_id in self.meshtastic_devices and self.meshtastic_devices[from_id]['uid']:
            uid = self.meshtastic_devices[from_id]['uid']
        elif not uid:
            uid = from_id

        now = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        stale = (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=1)).strftime("%Y-%m-%dT%H:%M:%SZ")

        event = Element('event', {'how': how, 'type': cot_type, 'version': '2.0',
                                  'uid': uid, 'start': now, 'time': now, 'stale': stale})

        SubElement(event, 'point', {'ce': '9999999.0', 'le': '9999999.0',
                                    'hae': str(self.meshtastic_devices[from_id]['last_alt']),
                                    'lat': str(self.meshtastic_devices[from_id]['last_lat']),
                                    'lon': str(self.meshtastic_devices[from_id]['last_lon'])})

        detail = SubElement(event, 'detail')
        if portnum == "TEXT_MESSAGE_APP" or (portnum == "ATAK_PLUGIN" and pb.HasField('chat')):
            return event, detail
        else:
            SubElement(detail, 'takv', {'device': self.meshtastic_devices[from_id]['hw_model'],
                                        'version': self.meshtastic_devices[from_id]['firmware_version'],
                                        'platform': 'Meshtastic', 'os': 'Meshtastic',
                                        'macaddr': self.meshtastic_devices[from_id]['macaddr'],
                                        'meshtastic_id': self.meshtastic_devices[from_id]['meshtastic_id']})
            SubElement(detail, 'contact',
                       {'callsign': self.meshtastic_devices[from_id]['long_name'], 'endpoint': 'MQTT'})
            SubElement(detail, 'uid', {'Droid': self.meshtastic_devices[from_id]['long_name']})
            SubElement(detail, 'precisionlocation', {'altsrc': 'GPS', 'geopointsrc': 'GPS'})
            SubElement(detail, 'status', {'battery': str(self.meshtastic_devices[from_id]['battery'])})
            SubElement(detail, 'track', {'course': '0.0', 'speed': '0.0'})
            SubElement(detail, '__group', {'name': self.meshtastic_devices[from_id]['team'],
                                           'role': self.meshtastic_devices[from_id]['role']})

            eud = EUD()
            eud.uid = uid
            eud.callsign = self.meshtastic_devices[from_id]['long_name']
            eud.device = self.meshtastic_devices[from_id]['hw_model']
            eud.os = "Meshtastic"
            eud.platform = "Meshtastic"
            eud.version = self.meshtastic_devices[from_id]['firmware_version']
            eud.team_role = self.meshtastic_devices[from_id]['role']
            eud.meshtastic_id = int(self.meshtastic_devices[from_id]['meshtastic_id'], 16)
            eud.meshtastic_macaddr = self.meshtastic_devices[from_id]['macaddr']

            with self.context:
                team = db.session.execute(db.session.query(Team).filter_by(name=self.meshtastic_devices[from_id]['team'])).scalar()
                if team:
                    eud.team_id = team.id

                try:
                    db.session.add(eud)
                    db.session.commit()
                except sqlalchemy.exc.IntegrityError:
                    db.session.rollback()
                    db.session.execute(sqlalchemy.update(EUD).where(EUD.uid == uid).values(**eud.serialize()))
                    db.session.commit()

        return event

    def position(self, pb, from_id, to_id, portnum):
        try:
            if portnum == "MAP_REPORT_APP" and pb.firmware_version != self.meshtastic_devices[from_id]['firmware_version']:
                try:
                    with self.context:
                        eud = self.db.session.execute(self.db.session.query(EUD).filter_by(uid=from_id)).first()[0]
                        eud.version = pb.firmware_version
                        eud.device = pb.hw_model
                        eud.callsign = pb.long_name
                        self.db.session.add(eud)
                        self.db.session.commit()
                        if from_id not in self.meshtastic_devices:
                            self.meshtastic_devices[from_id] = {'hw_model': '', 'long_name': '', 'short_name': '',
                                                                'macaddr': '',
                                                                'firmware_version': '', 'last_lat': "0.0",
                                                                'last_lon': "0.0",
                                                                'battery': 0, 'meshtastic_id': '',
                                                                'voltage': 0, 'uptime': 0, 'last_alt': "9999999.0",
                                                                'course': '0.0',
                                                                'speed': '0.0', 'team': 'Cyan', 'role': 'Team Member',
                                                                'uid': None}

                        self.meshtastic_devices[from_id]['firmware_version'] = pb.firmware_version
                        self.meshtastic_devices[from_id]['hw_model'] = mesh_pb2.HardwareModel.Name(pb.hw_model)
                        self.meshtastic_devices[from_id]['long_name'] = pb.long_name
                        self.meshtastic_devices[from_id]['short_name'] = pb.short_name
                except BaseException as e:
                    self.logger.error("Failed to update {}'s firmware version: {}".format(from_id, e))

            self.meshtastic_devices[from_id]['last_lat'] = pb.latitude_i * .0000001
            self.meshtastic_devices[from_id]['last_lon'] = pb.longitude_i * .0000001
            self.meshtastic_devices[from_id]['last_alt'] = pb.altitude
            if portnum == "POSITION_APP":
                self.meshtastic_devices[from_id]['course'] = pb.ground_track if pb.ground_track else "0.0"
                self.meshtastic_devices[from_id]['speed'] = pb.ground_speed if pb.ground_speed else "0.0"

            return self.cot(pb, from_id, to_id, portnum)
        except BaseException as e:
            self.logger.error("Failed to create CoT: {}".format(str(e)))
            self.logger.error(traceback.format_exc())
            return

    def text_message(self, pb, from_id, to_id, portnum):
        callsign = from_id
        if from_id in self.meshtastic_devices:
            callsign = self.meshtastic_devices[from_id]['long_name']

        chatroom = "All Chat Rooms"
        for meshtastic_device in self.meshtastic_devices:
            meshtastic_device = self.meshtastic_devices[meshtastic_device]
            if meshtastic_device['meshtastic_id'] == to_id:
                chatroom = meshtastic_device['uid']
                break

        if from_id in self.meshtastic_devices and self.meshtastic_devices[from_id]['uid']:
            from_uid = self.meshtastic_devices[from_id]['uid']
        else:
            from_uid = from_id

        message_uid = str(uuid.uuid4())
        event, detail = self.cot(pb, from_uid, chatroom, portnum, how='h-g-i-g-o', cot_type='b-t-f',
                                 uid="GeoChat.{}.{}.{}".format(from_uid, chatroom, message_uid))

        chat = SubElement(detail, '__chat',
                          {'chatroom': chatroom, 'groupOwner': "false", 'id': chatroom,
                           'messageId': message_uid, 'parent': 'RootContactGroup',
                           'senderCallsign': callsign})
        SubElement(chat, 'chatgrp', {'id': chatroom, 'uid0': from_uid, 'uid1': chatroom})
        SubElement(detail, 'link', {'relation': 'p-p', 'type': 'a-f-G-U-C', 'uid': from_uid})
        remarks = SubElement(detail, 'remarks', {'source': 'BAO.F.ATAK.{}'.format(from_uid),
                                                 'time': datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
                                                 'to': chatroom})

        remarks.text = pb.decode('utf-8', 'replace')

        return event

    def node_info(self, pb, from_id, to_id, portnum):
        if portnum == "ATAK_PLUGIN":
            uid = unishox2.decompress(pb.contact.device_callsign, len(pb.contact.device_callsign))
            self.meshtastic_devices[from_id]['uid'] = uid
            self.meshtastic_devices[from_id]['long_name'] = unishox2.decompress(pb.contact.callsign,
                                                                                len(pb.contact.callsign))
            self.meshtastic_devices[from_id]['short_name'] = uid[-4:]
            self.meshtastic_devices[from_id]['battery'] = pb.status.battery
            if pb.group.team != 0:
                self.meshtastic_devices[from_id]['team'] = atak_pb2.Team.Name(pb.group.team)
            if pb.group.role != 0:
                self.meshtastic_devices[from_id]['role'] = atak_pb2.MemberRole.Name(pb.group.role)
        else:
            hw_model = mesh_pb2.HardwareModel.Name(pb.hw_model)
            self.meshtastic_devices[from_id]['hw_model'] = hw_model if hw_model else ""
            self.meshtastic_devices[from_id]['long_name'] = str(pb.long_name) if pb.long_name else ""
            self.meshtastic_devices[from_id]['short_name'] = str(pb.short_name) if pb.short_name else ""
            self.meshtastic_devices[from_id]['macaddr'] = base64.b64encode(pb.macaddr).decode(
                'ascii') if pb.macaddr else ""

        return self.cot(pb, from_id, to_id, portnum)

    def telemetry(self, pb, from_id, to_id, portnum):
        if pb.HasField('device_metrics'):
            self.meshtastic_devices[from_id]['battery'] = pb.device_metrics.battery_level
            self.meshtastic_devices[from_id]['voltage'] = pb.device_metrics.voltage
            self.meshtastic_devices[from_id]['uptime'] = pb.device_metrics.uptime_seconds
        elif pb.HasField('environment_metrics'):
            self.meshtastic_devices[from_id]['temperature'] = pb.environment_metrics.temperature
            self.meshtastic_devices[from_id]['relative_humidity'] = pb.environment_metrics.relative_humidity
            self.meshtastic_devices[from_id]['barometric_pressure'] = pb.environment_metrics.barometric_pressure
            self.meshtastic_devices[from_id]['gas_resistance'] = pb.environment_metrics.gas_resistance
            self.meshtastic_devices[from_id]['voltage'] = pb.environment_metrics.voltage
            self.meshtastic_devices[from_id]['current'] = pb.environment_metrics.current
            self.meshtastic_devices[from_id]['iaq'] = pb.environment_metrics.iaq

    def atak_plugin(self, pb, from_id, to_id, portnum):
        self.node_info(pb, from_id, to_id, portnum)

        if pb.HasField('status'):
            self.meshtastic_devices[from_id]['battery'] = pb.status.battery

        if pb.HasField('pli'):
            self.meshtastic_devices[from_id]['last_lat'] = pb.pli.latitude_i * .0000001
            self.meshtastic_devices[from_id]['last_lon'] = pb.pli.longitude_i * .0000001
            self.meshtastic_devices[from_id]['last_alt'] = pb.pli.altitude
            self.meshtastic_devices[from_id]['course'] = pb.pli.course
            self.meshtastic_devices[from_id]['speed'] = pb.pli.speed
            return self.cot(pb, from_id, to_id, portnum)
        elif pb.HasField('chat'):
            self.logger.debug(
                "Got chat: {} {}->{}: {}".format(unishox2.decompress(pb.chat.to, len(pb.chat.to)), from_id, to_id,
                                                 unishox2.decompress(pb.chat.message, len(pb.chat.message))))

            chatroom = unishox2.decompress(pb.chat.to, len(pb.chat.to))
            message_uid = str(uuid.uuid4())

            from_uid = sender_callsign = from_id
            if from_uid in self.meshtastic_devices:
                from_uid = self.meshtastic_devices[from_id]['uid']
                sender_callsign = self.meshtastic_devices[from_id]['long_name']

            uid = "GeoChat.{}.{}.{}".format(from_uid, chatroom, message_uid)

            event, detail = self.cot(pb, from_uid, to_id, portnum, how='h-g-i-g-o', cot_type='b-t-f', uid=uid)

            chat = SubElement(detail, '__chat',
                              {'chatroom': 'All Chat Rooms', 'groupOwner': "false", 'id': chatroom,
                               'messageId': message_uid, 'parent': 'RootContactGroup',
                               'senderCallsign': sender_callsign})
            SubElement(chat, 'chatgrp', {'id': chatroom, 'uid0': from_uid, 'uid1': chatroom})
            SubElement(detail, 'link', {'relation': 'p-p', 'type': 'a-f-G-U-C', 'uid': from_uid})
            remarks = SubElement(detail, 'remarks', {'source': 'BAO.F.ATAK.{}'.format(from_uid),
                                                     'time': datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
                                                     'to': chatroom})
            remarks.text = unishox2.decompress(pb.chat.message, len(pb.chat.message))

            return event

    def protobuf_to_cot(self, pb, from_id, to_id, portnum, meshtastic_id):
        self.logger.debug(from_id + " " + to_id + " " + portnum + " " + meshtastic_id)
        event = None

        if from_id not in self.meshtastic_devices:
            self.meshtastic_devices[from_id] = {'hw_model': '', 'long_name': '', 'short_name': '', 'macaddr': '',
                                                'firmware_version': '', 'last_lat': "0.0", 'last_lon': "0.0",
                                                'battery': 0, 'meshtastic_id': meshtastic_id,
                                                'voltage': 0, 'uptime': 0, 'last_alt': "9999999.0", 'course': '0.0',
                                                'speed': '0.0', 'team': 'Cyan', 'role': 'Team Member', 'uid': None}

        if portnum == "MAP_REPORT_APP" or (portnum == "POSITION_APP" and pb.latitude_i):
            event = self.position(pb, from_id, to_id, portnum)
        elif portnum == "NODEINFO_APP":
            event = self.node_info(pb, from_id, to_id, portnum)
        elif portnum == "TEXT_MESSAGE_APP":
            event = self.text_message(pb, from_id, to_id, portnum)
        elif portnum == "ATAK_PLUGIN":
            event = self.atak_plugin(pb, from_id, to_id, portnum)
        elif portnum == "TELEMETRY_APP":
            self.telemetry(pb, from_id, to_id, portnum)

        try:
            if event:
                uid = self.meshtastic_devices[from_id]['uid']
                if not uid:
                    uid = from_id
                message = json.dumps({'uid': uid, 'cot': tostring(event).decode('utf-8')})
                if portnum == "TEXT_MESSAGE_APP":
                    try:
                        if to_id == "all":
                            self.rabbit_channel.basic_publish(exchange='chatrooms', routing_key='All Chat Rooms',
                                                              body=message,
                                                              properties=pika.BasicProperties(expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")))
                        else:
                            for meshtastic_device in self.meshtastic_devices:
                                meshtastic_device = self.meshtastic_devices[meshtastic_device]
                                if meshtastic_device['meshtastic_id'] == to_id:
                                    self.rabbit_channel.basic_publish(exchange='dms',
                                                                      routing_key=meshtastic_device['uid'],
                                                                      body=message,
                                                                      properties=pika.BasicProperties(expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")))
                    except BaseException as e:
                        self.logger.error("Failed to publish chat message: {}".format(e))
                elif portnum == "ATAK_PLUGIN" and pb.HasField('chat'):
                    try:
                        to = unishox2.decompress(pb.chat.to, len(pb.chat.to))
                        if to in self.meshtastic_devices:
                            self.rabbit_channel.basic_publish(exchange='dms', routing_key=to, body=message)
                        else:
                            self.rabbit_channel.basic_publish(exchange='chatrooms',
                                                              routing_key=to,
                                                              body=message,
                                                              properties=pika.BasicProperties(expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")))
                    except BaseException as e:
                        self.logger.error("Failed to publish chat message to {}: {}".format(
                            unishox2.decompress(pb.chat.to, len(pb.chat.to)), e))
                        self.logger.error(traceback.format_exc())
                else:
                    self.rabbit_channel.basic_publish(exchange='groups', routing_key=f"{self.context.app.config.get('OTS_MESHTASTIC_GROUP')}.OUT", body=message,properties=pika.BasicProperties(expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")))
                    self.rabbit_channel.basic_publish(exchange='cot', routing_key='', body=message, properties=pika.BasicProperties(expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")))
        except BaseException as e:
            self.logger.error(str(e))
            self.logger.error(traceback.format_exc())



"""Removed unique constraint on ATAK plugin names

Revision ID: eaf2af6f6b22
Revises: 8d6f811f4fd7
Create Date: 2025-11-08 16:28:41.700279

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'eaf2af6f6b22'
down_revision = '8d6f811f4fd7'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table('packages', schema=None) as batch_op:
            batch_op.drop_constraint(batch_op.f('packages_package_name_key'), type_='unique')
    except BaseException as e:
        print(e)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.create_unique_constraint(batch_op.f('packages_package_name_key'), ['package_name'])

    # ### end Alembic commands ###



from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, BooleanField


class MediaMTXPathConfig(FlaskForm):
    def __init__(self, formdata, **kwargs):
        super().__init__(formdata, **kwargs)

    name = StringField()
    source = StringField()
    sourceFingerprint = StringField()
    sourceOnDemand = BooleanField(default=True)
    sourceOnDemandStartTimeout = StringField()
    sourceOnDemandCloseAfter = StringField()
    maxReaders = IntegerField()
    srtReadPassphrase = StringField()
    fallback = StringField()
    record = BooleanField(default=False)
    recordPath = StringField()
    recordFormat = StringField()
    recordPartDuration = StringField()
    recordSegmentDuration = StringField()
    recordDeleteAfter = StringField()
    publishUser = StringField()
    publishPass = StringField()
    #publishIPs
    readUser = StringField()
    readPass = StringField()
    #readIPs
    overridePublisher = BooleanField()
    srtPublishPassphrase = StringField()
    rtspTransport = StringField()
    rtspAnyPort = BooleanField()
    rtspRangeType = StringField()
    rtspRangeStart = StringField()
    sourceRedirect = StringField()
    rpiCameraCamID = IntegerField()
    rpiCameraWidth = IntegerField()
    rpiCameraHeight = IntegerField()
    rpiCameraHFlip = BooleanField()
    rpiCameraVFlip = BooleanField()
    rpiCameraBrightness = IntegerField()
    rpiCameraContrast = IntegerField()
    rpiCameraSaturation = IntegerField()
    rpiCameraSharpness = IntegerField()
    rpiCameraExposure = StringField()
    rpiCameraAWB = StringField()
    rpiCameraDenoise = StringField()
    rpiCameraShutter = IntegerField()
    rpiCameraMetering = StringField()
    rpiCameraGain = IntegerField()
    rpiCameraEV = IntegerField()
    rpiCameraROI = StringField()
    rpiCameraHDR = BooleanField()
    rpiCameraTuningFile = StringField()
    rpiCameraMode = StringField()
    rpiCameraFPS = IntegerField()
    rpiCameraIDRPeriod = IntegerField()
    rpiCameraBitrate = IntegerField()
    rpiCameraProfile = StringField()
    rpiCameraLevel = StringField()
    rpiCameraAfMode = StringField()
    rpiCameraAfRange = StringField()
    rpiCameraAfSpeed = StringField()
    rpiCameraLensPosition = IntegerField()
    rpiCameraAfWindow = StringField()
    rpiCameraTextOverlayEnable = BooleanField()
    rpiCameraTextOverlay = StringField()
    runOnInit = StringField()
    runOnInitRestart = BooleanField()
    runOnDemand = StringField()
    runOnDemandRestart = BooleanField()
    runOnDemandStartTimeout = StringField()
    runOnDemandCloseAfter = StringField()
    runOnUnDemand = StringField()
    runOnReady = StringField()
    runOnReadyRestart = BooleanField()
    runOnNotReady = StringField()
    runOnRead = StringField()
    runOnReadRestart = BooleanField()
    runOnUnread = StringField()
    runOnRecordSegmentCreate = StringField()
    runOnRecordSegmentComplete = StringField()

    def serialize(self):
        return_value = {}
        for field in self._fields:
            if field != 'csrf_token':
                return_value[field] = self._fields[field].data
        return return_value



import pathlib
from dataclasses import dataclass
from urllib.parse import urlparse

from flask import request, current_app as app

from opentakserver.extensions import db
from sqlalchemy import Integer, String, ForeignKey, Boolean, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class VideoRecording(db.Model):
    __tablename__ = 'video_recordings'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    segment_path: Mapped[str] = mapped_column(String(255), unique=True)
    path: Mapped[str] = mapped_column(String(255), ForeignKey("video_streams.path"))
    in_progress: Mapped[bool] = mapped_column(Boolean)
    start_time: Mapped[DateTime] = mapped_column(DateTime, nullable=True)
    stop_time: Mapped[DateTime] = mapped_column(DateTime, nullable=True)
    duration: Mapped[int] = mapped_column(Integer, nullable=True)
    width: Mapped[int] = mapped_column(Integer, nullable=True)
    height: Mapped[int] = mapped_column(Integer, nullable=True)
    video_codec: Mapped[str] = mapped_column(String(255), nullable=True)
    video_bitrate: Mapped[int] = mapped_column(Integer, nullable=True)
    audio_codec: Mapped[str] = mapped_column(String(255), nullable=True)
    audio_bitrate: Mapped[int] = mapped_column(Integer, nullable=True)
    audio_samplerate: Mapped[int] = mapped_column(Integer, nullable=True)
    audio_channels: Mapped[int] = mapped_column(Integer, nullable=True)
    file_size: Mapped[int] = mapped_column(Integer, nullable=True)
    video_stream = relationship("VideoStream", back_populates="recordings")

    def serialize(self):
        return {
            'segment_path': self.segment_path,
            'path': self.path,
            'in_progress': self.in_progress
        }

    def to_json(self):
        with app.app_context():
            url = urlparse(request.url_root)
            protocol = url.scheme
            hostname = url.hostname
            port = url.port
            if not port and protocol == 'https':
                port = 443
            elif not port and protocol == 'http':
                port = 80

            return {
                'id': self.id,
                'segment_path': self.segment_path,
                'path': self.path,
                'in_progress': self.in_progress,
                'start_time': iso8601_string_from_datetime(self.start_time) if self.start_time else None,
                'stop_time': iso8601_string_from_datetime(self.stop_time) if self.stop_time else None,
                'duration': self.duration,
                'filename': pathlib.Path(self.segment_path).name,
                'width': self.width,
                'height': self.height,
                'video_codec': self.video_codec,
                'video_bitrate': self.video_bitrate,
                'audio_codec': self.audio_codec,
                'audio_bitrate': self.audio_bitrate,
                'audio_samplerate': self.audio_samplerate,
                'audio_channels': self.audio_channels,
                'file_size': self.file_size,
                'thumbnail': f"{protocol}://{hostname}:{port}/api/videos/thumbnail?path={self.path}&recording={pathlib.Path(self.segment_path).name}"
            }



from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import String, Float, LargeBinary
from sqlalchemy.orm import Mapped, mapped_column


@dataclass
class APSchedulerJobs(db.Model):
    __tablename__ = "apscheduler_jobs"

    id: Mapped[str] = mapped_column(String(255), primary_key=True)
    next_run_time: Mapped[float] = mapped_column(Float, nullable=True, index=True)
    job_state: Mapped[bytes] = mapped_column(LargeBinary, nullable=False)



import flask_security
import unicodedata
from flask_security import UsernameUtil
from opentakserver.extensions import logger


class UsernameValidator(UsernameUtil):

    def __init__(self, app):
        super().__init__(app)
        self.app = app

    def check_username(self, username: str) -> str | None:
        """
        Allow letters, numbers, underscores and periods in usernames
        """
        for character in username:
            if unicodedata.category(character)[0] not in ["L", "N"] and character != "_" and character != ".":
                return flask_security.utils.get_message("USERNAME_DISALLOWED_CHARACTERS")[0]
        return None



"""Added CasEvac fields

Revision ID: 3a7ea7ead17b
Revises: d68964dab66b
Create Date: 2025-10-03 04:02:44.032178

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3a7ea7ead17b'
down_revision = 'd68964dab66b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('casevac', schema=None) as batch_op:
        batch_op.add_column(sa.Column('convenience', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('urgent_surgical', sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('casevac', schema=None) as batch_op:
        batch_op.drop_column('urgent_surgical')
        batch_op.drop_column('convenience')

    # ### end Alembic commands ###



from flask_wtf import FlaskForm
from flask_wtf.file import FileRequired, FileAllowed
from wtforms import StringField, FileField, BooleanField
from wtforms.validators import Optional, DataRequired
from opentakserver.functions import false_values


class PackageForm(FlaskForm):
    platform = StringField(default="Android")
    plugin_type = StringField(default="plugin")
    apk = FileField(validators=[FileRequired(), FileAllowed(['apk'])])
    icon = FileField(validators=[Optional()])
    description = StringField(validators=[Optional()])
    install_on_enrollment = BooleanField(false_values=false_values)
    install_on_connection = BooleanField(false_values=false_values)
    atak_version = StringField(validators=[Optional()])


class PackageUpdateForm(FlaskForm):
    package_name = StringField(validators=[DataRequired()])
    install_on_enrollment = BooleanField(false_values=false_values)
    install_on_connection = BooleanField(false_values=false_values)
